//Program K38 EvoFull
/*
 * We implement here 
 * a full fledged evolutionary process
 * to propose a solution to the problem of
 * fitting a tree to a distance matrix.
 * 
 * 
 * We generate a random tree topology 
 * and attach random integer lengths to links.
 * In that way we get the target tree, T. 
 * Next, we find the associated distance matrix D  to T.
 * Our problem is:
 * 
 * Find tree t such that the absolute distance between
 * matrix D and the associated matrix to t is a minimum.
 * Absolute distance between M and N = sum |Mij-Nij|.
 * This distance defines the FITNESS of each tree.
 * It decreases as good performance augments.
 * The best possible fitness is zero.
 * 
 * We run the next test:
 * If we begin with a tree T, compute D, and then t,
 * we expect to return with the initial tree: t = T.
 *
 * 
 * To associate a distance matrix to a tree, 
 * we calculate the distance in within 
 * each pair of nodes (leaves + inner nodes).
 * To that aim, we begin with the first node and 
 * walk over the tree until the second node
 * and add the lengths of all traversed links.
 * The resulting number is the  distance.
 *  
 *  
 * Formats = Link and Newick's notations. And variants. 
 * 
 * An automatic halting option has been added to be applied
 * if perfect matching is found.
 * The program accepts for input 
 * a distance matrix with integer values.    
 * A test for being a distance matrix is run.
 * 
 * To find the tree that best fit D, 
 * we  use evolution as follows:
 * 
 * 1) Initialization:
 * 		We generate N topologies  on the same
 * 		set of leaves. By adding   lengths to links,
 * 		every topology gets represented by a tree.
 *		The fitness of each tree
 *		is calculated and the best is selected.
 * 		If the fitness of the best is zero, 
 * 		the program halts and the tree is reported.
 * 
 * 2) Evolution is composed of rounds of
 *   	 cloning according to fitness, 
 *   	 random mutation and recombination.
 *       After each generation, fitness is checked 
 *   	 for accomplishment. 
 * 
 * 
 * Mutation is Lamarckian: 
 * it always produces viable mutants. 
 * Topologies as well as links can mutate.
 * 
 * A mutation results from deleting 
 * a given number of leaves
 * to reinsert them anew in 
 * other places chosen at random.
 * This creates a mutant topology.  
 * To mutate length of links:
 * Lengths are read from tree and 
 * mutated separately.
 * Then, they are reinserted.
 * 
 * 
 * Recombination is lamarckian: it is authorized only
 * when it produces viable recombinants.
 * All their subtopologies are calculated 
 * and packed in arrays.
 * The sets of leaves and of inner nodes 
 * of each subtopology are found.
 * Subtopologies are compared by pairs,
 * one from each topology. 
 * If they happen to have the same set of leaves
 * and of inner nodes, recombination is authorized:
 * the  subtrees that correspond with 
 * the subtopologies are swapped.
 * 
 * If more than one pair of subtopologies authorize
 * recombination, one is chosen at random.
 * A hashtable is used to memorize those pairs that do it.
 * Original topologies serves as subtopologies 
 * that authorize recombination but they do not count.
 *
 * 
 */



import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import java.awt.*;



import javax.swing.JComponent;


public class EvoFull extends JPanel 
                          implements ActionListener 
{
 
 private static final long serialVersionUID = 1L;

 //The program accepts a distance matrix as input
 //else
 //it generates a random tree, calculates the distance
 //matrix associated to leaves and posits it 
 //as target matrix.
 private static boolean inputMatrix = false;
 

 //Input distance matrix: 
 //Example 5.4, pag 99, Isaev (2001) 
 //Introduction to Mathematical Methods in Bioinformatics
 //Springer
 private static double targetMatrix[][] 
 = {{0,8,3,14,10,12},
   {8,0,9,10,6,8},
   {3,9,0,15,11,13},
   {14,10,15,0,10,8},
   {10,6,11,10,0,8},
   {12,8,13,8,8,0}};

 /*
 private static double targetMatrix5[][] 
		 = {{0,8,3,14,10},
		   {8,0,9,10,6},
		   {3,9,0,15,11},
		   {14,10,15,0,10},
		   {10,6,11,10,0}};

 private static double targetMatrix4[][] 
		 = {{0,8,3,14},
		   {8,0,9,10},
		   {3,9,0,15},
		   {14,10,15,0}};
		   
		   */

 //Number of leaves that must go in all trees > 3	
 private static	int numberOfLeaves;

 //Everything is built by recursion over this tree
 private static   String   seed = 
		    "((Aa,Ba),Ca)";
 
 //The best distance found in initialization	
 private static  int  distance0;


 private static  String leaves =  
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 private static String internalNodes = 
		 "abcdefghijklmnopqrstuvwxyz";
 
 //Next leaf to be grafted	
 private static  int nextLeafIndex = 2;
 //Next internal node to be inserted
 private static  int nextIntNodeIndex = 0;
 
 
 private static int maxNumberOfLeaves = leaves.length();
 
 
 


 //Number of leaves to be deleted and reinserted
 //to create a mutation
 private static int nUndoRedo = 0;


 private static boolean print1 = false;


 //This is the original topology:
 private static  String original = "";


 
 
 //===========Most important parameters =========== 
 
 //Number of leaves to be inserted to seed:
 //to get a tree with 4 leaves, add one leaf to seed
 private static  int nAdditionalLeaves = 3; 


 
 //Number of topologies = number of trees that
 //conform the evolving population
 private static  int N = 50;
 
 //Max number of generations to run evolution 
 static private int numberOfGenerations = 50000001;
 static private long generation = 0;

 //Mutation rate per site   per generation
 //in the lengths of links
 static private double    m = 0.5;



//Report Best tree every lapse generations
static private int lapse = 1000;

 //=================================================
 
 
 //Data for binary tree
 private static String[] Record =  new String[20*N];
 

 private static  int indexOfTopologies;
 private static String topologies[] = new String[N];
 private static String trees[] = new String[N];
 private static String newTrees[] = new String[N];
 
 //The distance to target of the best fitting tree 
 //of the given topology
 private static double[] bestFitting =  new double[N];
 private static String[] bestTreeOfTopology
                             = new String[N];
 
 //Last filled position  in Record[] 
 private static int lastUsedPosition;	
 private static String selectedTree = "";
 
 
  //DynamicTree is an inner service class that comes
  //at the end of this file.
  private static DynamicTree15 treePanel;
  private static String VISUALIZE_COMMAND = "Visualize";
  
   //Length of the link that glues a leaf to the tree
  private static int lengthsLeaves[]  = 
		                new int[maxNumberOfLeaves];
  //Length of the link that glues 
  //an inner node to another one
  private static int lengthsInnerLinks[]  = 
		               new int[maxNumberOfLeaves];
   
  
  private static Random r = new Random();
  //The control parameter of the random generator for int
  private static int maxLength = 10;
  
  //Paths from each leaf to node a
  private static String[] Paths =  
		               new String[maxNumberOfLeaves];
  
  private static String targetTree;
 
  
  
  //=======Evolution parameters============




 static private double distance[] = new double[N];
 static private double fitness[] = new double[N];



 //Index of the best
 static private int best;
 //The best tree for a given topology
 static private String bestTree = "";


 //=========Debugging help ===============

 private static boolean print  = false;

 private static boolean print2 = false;
 private static boolean print3 = false;
 private static boolean print4 = false;
 //Turn on print5 firstly
 private static boolean print5 = false;
 private static boolean print6 = false;
 private static boolean print7 = true;


 //==============

 private static String titleFrame = "";
 private static String title1 = "";
 private static String title2 = "";
 private static EvoFull newContentPane = new EvoFull();



  //=======Constructor of main class: GUI basics=====

  public EvoFull() 
  {
   super(new BorderLayout());
   JPanel instructionPanel = new JPanel(new GridLayout(0,1)); 
   JLabel instructions1 = new JLabel(
    	 "Select  a tree and click the Visualize button.");
 

   instructionPanel.setFocusable(true);
   instructionPanel.add(instructions1);
 
   add(instructionPanel, BorderLayout.PAGE_END);
   
   //A new node is added
   JButton visualizeButton = new JButton("Visualize");
   visualizeButton.setActionCommand(VISUALIZE_COMMAND);
   visualizeButton.addActionListener(this);
    
   // Create the components.
   treePanel = new DynamicTree15();    
   // Lay everything out.
   treePanel.setPreferredSize(new Dimension(700, 800));
   instructionPanel.add(visualizeButton);
   add(treePanel, BorderLayout.CENTER);
  }//End of constructor

  //Response to selecting visualization
  public void actionPerformed(ActionEvent e) 
   {
	String command = e.getActionCommand();
    if (VISUALIZE_COMMAND.equals(command)) 
	 {
      //Selected topology must be read
      selectedTree = treePanel.unveilTree();      
      System.out.println("\nSelected topology " +
      		"in Link notation  after unveilTree() = \n"  
    		  + selectedTree);
      /*Random lengths are added to topology
      generateLengths();
      //Lengths are added
      selectedTree = insertLengths(selectedTree);
      */
      int i =  findTheBest(selectedTree);
      selectedTree =  bestTreeOfTopology[i];
      double d = bestFitting[i];
      System.out.println( " Best fitting = \n " +
                                  selectedTree);
      System.out.println( " Best distance = \n "
    	  + d); 
      
      
      decoderLengths(selectedTree);    
	  //Translation from one format into another
      //output = a vector with binary addresses and contents
      double[][] M =  associatedMatrix(selectedTree); 
      printMatrix(M, "distance matrix associated to the " + 
      "best tree for this topology");
      //Selected tree is drawn
      title1 = "Best tree for the topology = " 
    		  + simplify(selectedTree);    
      title2 = "Distance to target = " + d;
      DrawBiTree12 a = newContentPane.new DrawBiTree12();
      a.createAndShowGUI();
      
	 } 
  }

//======================

  




public static void main(String[] args) 
{
   // Schedule a job for the event-dispatching thread:
   // creating and showing this application's GUI.
   javax.swing.SwingUtilities.invokeLater(new Runnable() 
    {
     public void run() 
     {
      //If input is a distance matrix:  
      if (inputMatrix)
      {
       printMatrix(targetMatrix, "Input Matrix");
       boolean isD = isMatrixDistanceMatrix(targetMatrix);
         if (isD) System.out.println("\nMatrix " +
       		"is a distance matrix"); 
         else 
          {
    	   System.out.println("Matrix " +
    	       		"is NOT a distance matrix"); 
    	   System.exit(0);
           }
         numberOfLeaves = targetMatrix.length;
         nAdditionalLeaves = numberOfLeaves-3;
    }
    else//if a random tree is generated -> Matrix
    {      	  
     numberOfLeaves =    3 + nAdditionalLeaves;
     String tree = randomTopology();
     generateLengths();
     targetTree = insertLengths(tree);
     System.out.println("Target tree: each link " +
    			"has its own length = \n" + targetTree);
     targetMatrix  =
   		     new double[numberOfLeaves][numberOfLeaves];
     targetMatrix = associatedMatrix(targetTree);
     printMatrix(targetMatrix, "Target Matrix");
    	
     double d = absDistance(numberOfLeaves,
    			              targetMatrix, targetMatrix); 
 	 
    	
     System.out.println( "\nDistance from target matrix " +
     		"to target = " + d);
    	 
    	
    	
    	
    	 //Draw Target tree: this malfunctions.
        /* 
         {
          titleFrame = "Target tree = " + tree;
           selectedTree = targetTree;
           DrawBiTree12 b = newContentPane.new DrawBiTree12();        
           b.createAndShowGUI();
           }
          */ 
     }
    createAndShowGUI();

    System.out.println("\nDefinition: \nDistance to target = " +
        		 " Absolute distance between" +
        		" target matrix  and \n" +
        		"the distance matrix associated to " +
        		"the tree with lengths of links that \n" +
        		"is the best fitting  of the topology.\n" +
        		"Perfect fitting: distance = 0 \n");
     System.out.println("\nNumber of Generations = " 
        		   + numberOfGenerations);
     if (print7)
     
     System.out.println("Number of leaves = " 
                         + numberOfLeaves +
                          "\nMutation rate = "   + m +
    		             "\nPopulation size = " + N);
     
     initialization();
     evolution();
     
     }
    });
  }




//Test for being a distance matrix. 
//MaxLenght = max value among entries is calculated.
private static boolean 
    isMatrixDistanceMatrix(double[][] M)
{
 int l = M.length;
 boolean isD = true;
 //Is M(u,u) = 0?
 for(int u = 0; u < l; u++)

	if (M[u][u] ==  0) ;
	 else 
	  {
		isD = false;
		return isD;
	  }

 //Is M(u,v) = M(v,u)?
 for(int u = 0; u < l; u++)
	 for(int v = u+1; v < l; v++)
		if (M[u][v] == M[v][u] ) ;
		 else 
			{
			 isD = false;
			 return isD;
			}

 //Is M(u,v)>0 for u not v?
 for(int u = 0; u < l; u++)
	 for(int v = u+1; v < l; v++)
		if (M[u][v] >  0) ;
		 else 
			{
			 isD = false;
			 return isD;
			}

	

 //Is M(u,v) < M(u,w) + M(w,v)?
 for(int u = 0; u < l; u++)
 for(int v = 0; v < l; v++)
 for(int w = 0; w < l; w++)
	   
		if (M[u][v] <= M[u][w]+M[w][v]  ) ;
		 else 
			{
			 isD = false;
			 return isD;
			}	
 double m = 0;
 for(int u = 0; u < l; u++)
	 for(int v = u+1; v < l; v++)
		 if (M[u][v] > m ) m = M[u][v];
 maxLength = (int) m + 1;
 return true; 
}



//Simplification from Link notation to Newick's notation:
//(((AJ,DJ)I,BI),CI) -> (((A,D),B),C)	 	 
private static String linkToNewick(String newTree)
{
 String tree = "";
 //length of string newTree
 int l = newTree.length();
 for(int i = 0; i < l; i++ )
	  {
	   //newTree is scanned char by char	 
	  char c = newTree.charAt(i);
	  //Is c an internal node to be deleted?
	  if (charIsInternalNode(c));
	   else //add char c to output
	    tree = tree + c;
	  }
	 return tree;	 
}



private static String generateTopology()
 {
  //Next leaf to be grafted	
  nextLeafIndex = 2;
  //Next internal node to be inserted
  nextIntNodeIndex = 0;
  //A tree topology is generated	
  String topology = seed;
  for(int i = 0; i < nAdditionalLeaves; i++ )	 
	 topology = generateChild(topology);
  return topology;
 }
 

//Random lengths are added to a tree in Link notation 
private static void makeLengths()
{
//Lengths with digit values are generated
for(int i = 0; i < numberOfLeaves; i++)
 {
	 //Zero lengths are not allowed 
	 lengthsLeaves[i] = 1 + r.nextInt(maxLength - 1); 
	 lengthsInnerLinks[i] = 1 +  r.nextInt(maxLength - 1);
 }
} 

//The population is initialized with nTop topologies.
//Lengths are inserted to get nTop trees
//Fitness is assessed:
//Has the solution been guessed?
private static void initialization()
{
 System.out.println("\n====" +"INITIALIZATION : ");
 if (print5)
  System.out.println("\nSeed = " + seed
				 + " -> "    + linkToNewick(seed));

 if (print7)
  System.out.println( N  
		   + " random topologies " +
  		"and their trees are generated \n");

 //nTop topologies and trees are generated 
 for(int i = 0; i < N; i++)
 {
  String topo1 = generateTopology();
  topologies[i] = topo1;
  String tree = insertLengths(topo1);
  makeLengths();
  //Lengths of links are added to tree topology
  tree = insertLengths(topo1);
  trees[i] = tree;
  if (print6)
  {
	System.out.println("\n" +  i + "   " + topo1 + " -> " 
	          + linkToNewick(topo1));
	System.out.println("Tree1 = " + tree);
  }
 
 }
 assessFitnesses();
 scaleFitnesses();
}


//=====================================
//============= EVOLUTION =============
//=====================================


//Fitness is calculated for each tree
//Fitness  =
//Distance between target matrix and 
//the associated distance matrix to tree is calculated.
//The minimal distance is found: 
//this renders the fittest tree.
//Has the solution been guessed?
private static double assessFitnesses()
{
 if (print6)
  System.out.println(  "\n=====Fitnesses are assessed");
 
 String tree = "";
 double d = 0;
for(int i = 0; i < N; i++)
{

 tree = trees[i];
 decoderLengths(tree);
 double M[][] = distanceMatrix(tree);
 //Absolute distance from the associated distance matrix
 //of the tree
 //to target matrix is evaluated
 d = absDistance(numberOfLeaves,targetMatrix, M); 
 distance[i] = d;

 
 if (print6)
  {
   System.out.println( "\nTree = \n" + trees[i]);
   System.out.println( "\nDistance to target = " + d);
  }


if (print5)
 {
  printMatrix(M, "Distance matrix associated to tree");
  printMatrix(targetMatrix, "Target matrix");
}

//If solution has been found, halt.
if (d == 0)
 {
  System.out.println( "\n \nSOLUTION HAS BEEN FOUND: \n" +
			"at generation " + generation);
  System.out.println( " Distance to target = " + d
		   + " \nSolution tree = " + tree
		   + " \nTarget tree   = " + targetTree
		   + " \nInitial distance = " + distance0);
printMatrix(M, "Distance matrix associated to tree");
printMatrix(targetMatrix, "Target matrix");
if (d == 0) System.exit(0);
if (print5) print5 = false;			  	
 }
}//End of for(int i...




//The fittest individual minimizes distance function
double min = distance[0];
best = 0;
for(int i = 1; i < N; i++)
{
 if (distance[i] < min )
  {
   min = distance[i];
   best = i;
  }
}

 //Best tree 
 bestTree = trees[best];
 if (generation == 0 ) distance0 = (int) min;
 if (print7)
 System.out.println("\n BestTree  = \n" +  bestTree    
		           + "\nDistance = " + min 
		           + "\ntarget tree = \n" + targetTree 
		           + "\nInitial distance = " + distance0);
 return min;
}



//Fitness is rescaled to take values
//in within 0(the worst) and 1(the best)
private static void scaleFitnesses()
{
	if (print6)
 System.out.println(  "\n====Fitnesses are rescaled: " +
 		"0(the worst) and 1(the best)");
double maxD = distance[0];
double minD = distance[0];
double b = 1 ;
//Min and max distances are found
for(int i = 1; i < N; i++)
 {
	if ( distance[i] > maxD ) maxD =  distance[i];
	if ( distance[i] < minD ) minD =  distance[i];
 }
	 
if (minD == maxD) //All   equally fit
 {
  for(int i = 0; i < N; i++)
	 fitness[i] = 1;
 }
else
 {
	 //Distance to target determine fitness:
	//fitness[maxD] = 0;  	 fitness[minD] = 1;
	 for(int i = 0; i < N; i++)
	 {  
	  b = 1;
	  b = b/(minD - maxD);
	  double a = -b*maxD;
	  fitness[i] = a + b * distance[i];	 
	}
 }
if (print6)
 for(int i = 0; i < N; i++)
 {
	 System.out.println("Individual = "
             + i + " has distance " +
          distance[i] + " and fitness " + fitness[i]);
 }
}

//Probability of cloning is proportional to fitness.
//N new individuals are formed. 
private static void selectiveCloning()
{
 if (print6)
  System.out.println(  "\n====Selective cloning");
//The best is cloned
 newTrees[0] = trees[best];
 if (print6)
	 System.out.println("Individual = "
                      + best + " with fitness " +
			 fitness[best] + "-> new individual " + 0
			 + " = "  + newTrees[0]  );	
//N-1 children are sampled
for(int i = 1; i < N; i++)
{
 //flag
 boolean go = true;
 while (go)
  {
	 //An individual is chosen at random  
	 int k = r.nextInt(N);
	 //Probability of cloning is proportional to fitness. 
	 double q = r.nextDouble();	 
	 if ( q < fitness[k]) //make a copy of k
	  {	    
		 newTrees[i] = trees[k];		  
		 if (print6)
			 System.out.println("Individual = "
		                      + k + " with fitness " +
					 fitness[k] + "-> new individual " + i
					 + " = "  + newTrees[i]  );		 
		 go = false;//end task		  
	  }
  }

}//End of for //N-1 children are sampled
}

//============Mutation============





//Detect internal nodes, different than leaves
private static boolean charIsInternalNode(char c)
{
 String s = "" + c;	 
 if (internalNodes.contains(s)) return true;
 else return false;
}



//Simplification from Link notation to Newick's notation:
// (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)
private static String simplify(String newTopo)
{
String topo = "";
//length of string newTopo
int l = newTopo.length();
for(int i = 0; i < l; i++ )
 {
  //newTopo is scanned char by char	 
 char c = newTopo.charAt(i);
 //Is c an internal node to be deleted?
 if (charIsInternalNode(c));
  else //add char c to output
   topo = topo + c;
 }
return topo;	 
}

//Rule 2 is used to generate a new topology:
//(S)I -> ((S)N,FN)I.
//S = subtopology, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string parentTopo
private static String useRule2(String parentTopo, int i, 
       char F, char N)
{
 String newTopo = "";
 char I = parentTopo.charAt(i);
 if (print1)
  System.out.println("i = " + i + " I = " + I);
 //Subtopology S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = i;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
   String a = "" + parentTopo.charAt(position);
  
   if (print1)
   System.out.println("a = " + a + " p = " + pr);
   if (a.equals(pr) ) 
    {
	 counter++;
	 S = a + S;
    }
   else go = false;
  }//End of while
 if (print1)
 System.out.println("counter = " + counter);
 //The complete subtopology is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String a = "" + parentTopo.charAt(position);
   if (a.equals(pr))   counter++;
   if (a.equals(pl))  counter--;
   S = a + S;
   if (print1)
    System.out.println("Subtopo = " + S);
   position = position -1;
   }
 //Replacement rule 2:
 //((S)n,Fn)i -> (S)i
 //where   F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =   S  + I; 
 String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
 //Rule 1 is executed
 newTopo = parentTopo.replace(SnFni, replacement);		
 //If I links to a complex structure
 return newTopo;	 
}

//From parentTopo, a topology in Newick's notation,
//a Child  topology is generated by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String parentTopo) 
{
	
 nextLeafIndex++;
 char nextLeaf= leaves.charAt(nextLeafIndex);
 //Next internal node to be inserted
 nextIntNodeIndex++;
 char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
	
	
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new topology generated by replacement
 String newTopo = "";
 //length of parent topology
 int l = parentTopo.length();
 //counter of links = 0;
 int counter = 0;
 //Links are counted
 for(int i = 0; i < l; i++ )
  {
   //parentTopo is scanned char by char	 
   char I = parentTopo.charAt(i);
   //Is I an internal node that denotes a link?
   if (charIsInternalNode(I)) counter++;	
 }
 //A link is chosen 
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
 {
   //parentTopo is scanned char by char	 
	char I = parentTopo.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = parentTopo.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTopo = useRule2(parentTopo, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTopo = parentTopo.replace(link, replacement);		
		 }
		if (print2)
		System.out.println(newTopo + " -> " 
		                       + simplify(newTopo));			

	  
		 i = l;
	   }
	  counter++;	
	} 
 } 
return newTopo;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply the antirule1:
//(An,Fn)i  -> Ai
private static String  antiRule1(String topo, 
		int positionLeaf, int positionNode)
{
 if (print2)
  System.out.println("AntiRule1");
 int beginIndex = positionLeaf - 4;
 int endIndex = positionNode + 3;
 String AnFni = topo.substring(beginIndex, endIndex);
 if (print2)
  System.out.println("Anfni = " + AnFni);
 String Ai = "" + topo.charAt(beginIndex+1)
               + topo.charAt(endIndex-1);
 if (print2)
  System.out.println("Ai = " + Ai);
 topo = topo.replace(AnFni, Ai);		
	return topo;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply the antirule2:
//((S)n,Fn)i -> (S)i
//positionNode refers to the last appearing of node n
private static String  antiRule2(String topo, 
		int positionLeaf, int positionNode)
{
 if (print2)
	System.out.println("AntiRule2");
 //Leaf to be deleted	
 String leaf = "" + leaves.charAt(nextLeafIndex);
 if (print2)
  System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
	            + internalNodes.charAt(nextIntNodeIndex);	
 int firstPosition = topo.indexOf(innerNode);
 String newTopo = "";
 char i = topo.charAt(positionNode+2);
 if (print2)
  System.out.println("FirstPosition = " + 
                     firstPosition + " i = " + i);
 //Subtopology S of link (S)I must be read 
 String S = "";
 //Counter of left parenthesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = firstPosition;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
  {
   position = position -1;
   String as = "" + topo.charAt(position);
  
   if (print2)
    System.out.println("a = " + as + " p = " + pr);
   if (as.equals(pr) ) 
    {
	 counter++;
	 S = as + S;
    }
   else go = false;
  }//End of while
 if (print2)
 System.out.println("counter = " + counter);
 //The complete subtopology is captured:
 //Parentheses must balanced
 while (counter > 0)
  {   
   String as = "" + topo.charAt(position);
   if (as.equals(pr))   counter++;
   if (as.equals(pl))  counter--;
   S = as + S;
   if (print2)
    System.out.println("Subtopo = " + S);
   position = position -1;
  }
 String n = innerNode;
 //Replacement antiRule 2:
 //((S)n,Fn)i -> (S)i
 //where A and F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =  "("+ S  + n + "," + leaf + n + ")" + i ; 
 String	Si = S + i;
 if (print2)
  System.out.println("\nSnFni = " + SnFni + "\n Si = " + Si);
 //Rule 1 is executed
 newTopo = topo.replace(SnFni, Si);		
 return newTopo;
}



//The last inserted  leaf is deleted from topology.
private static String deleteLastLeaf(String topology) 
{
 //Leaf to be deleted	
 String leaf = "" + leaves.charAt(nextLeafIndex);
 if (print2)
 System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
           + internalNodes.charAt(nextIntNodeIndex);
 if (print2)
 System.out.println("Accompanying inner  node = " 
                   + innerNode);		
 //new topology = output
 String newtopology = "";
 //length of  topology as string
 int l = topology.length();
 //Last positions are found
 int positionLeaf = -1;
 int positionNode = -1;  
 //topology is scanned char by char
 for(int i = 0; i < l; i++ )
  {
	 String c = "" + topology.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
  }
 if (print2)
  {
   System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
   System.out.println("Position of " + innerNode + 
         " = " + positionNode);
  }
 int firstPosition = topology.indexOf(innerNode);
 //Check for the char before firstPosition	
 char a = topology.charAt(firstPosition - 1);
 //If I links to a complex structure
 if (a == ')')
 newtopology = antiRule2(topology,positionLeaf,positionNode);
 else  //Apply antiRule1 for simple cases
	  newtopology = antiRule1(topology,positionLeaf,positionNode);
 nextLeafIndex--;
 nextIntNodeIndex--;
 return newtopology;	
}



//==========================

//Lengths are deleted from newTree
private static String deleteLengths(String newTree)
{
 String tree = "";
 String numeric = ".:0123456789";
 //length of string newTree
 int l = newTree.length();
 for(int i = 0; i < l; i++ )
  {
   //newTree is scanned char by char	 
   char c = newTree.charAt(i);
   String s = "" + c;
   //Is c numeric: delete it!
   if (numeric.contains(s)); //add char c to output
	  else tree = tree + c;
  }
 return tree;	 
}





//Mutation:
//some leaves are deleted and then reinserted 
//in random order.
//Deleting can happen in either of two ways:
//Reversing the order of synthesis else
//in random order.
public static String mutation(String topology, 
		                       int nAdditionalLeaves)
{
 //Mutation has two stages:
 //deletion of some leaves that undoes generating work + 
 //re-insertion in places chosen at random.

 //Number of leaves to be deleted and then reinserted
 if (nAdditionalLeaves == 1) nUndoRedo = 1;
  else
   nUndoRedo = r.nextInt(nAdditionalLeaves-1) + 1;


 if (print2)
  System.out.println("Number of leaves to" +
		" be deleted and then reinserted = " +
		nUndoRedo); 

 //Some leaves are deleted: 
 //reversed order to that of synthesis.
 if (print1)
  System.out.println("Some leaves are deleted");
 for(int i = 0; i < nUndoRedo; i++ )	 
	 {
	   //Last inserted leaf is deleted.
	   topology = deleteLastLeaf(topology);
	   if (print1)
	    System.out.println(topology);
	 }
 //deleted leaves are reinserted
 if (print1)
  System.out.println("Deleted leaves are reinserted");
 for(int i = 0; i < nUndoRedo; i++ )	 
  {
	  topology = generateChild(topology);
  } 



 //Output
 String mutant = topology;
 if (print1)
  System.out.println("\nOriginal and mutant topologies = \n" 
		  + original + "\n" + mutant);
 if (print1)
 if (original.equals(mutant))
			  System.out.println("Mutation made nothing");
	   else
			  System.out.println("\nMutation worked");	
 return mutant;
}


public static String mutation(String topology)
{
 //This original topology is next subject to mutation
 original = topology;
 String mutantTopology = topology;
 if (print2)
	System.out.println("\nOriginal   topology = \n" 
			                + original  );
 //Number of leaves to be deleted and then reinserted
 if (nAdditionalLeaves == 1) nUndoRedo = 1;
	else
 nUndoRedo = r.nextInt(nAdditionalLeaves-1) + 1;
 //nUndoRedo = 13;
	 
 //Mutation has two stages:
 //deletion of some leaves that undoes generating work + 
 //re-insertion in places chosen at random.
	 
 if (print2)
	  System.out.println("Number of leaves to" +
	  		" be deleted and then reinserted = " +
	  		nUndoRedo); 
 //Some leaves are deleted:
 if (print2)
	 System.out.println("Some leaves are deleted");
 for(int i = 0; i < nUndoRedo; i++ )	 
  {
	//Last inserted leaf is deleted.
	mutantTopology = deleteLastLeaf(mutantTopology);
	if (print2)
	   System.out.println(mutantTopology);
  }
 //Some leaves are reinserted
 if (print2)
 System.out.println("Deleted leaves are reinserted");
 for(int i = 0; i < nUndoRedo; i++ )	 
  {
   topology = generateChild(topology);
  }
 return mutantTopology;	
}



//Random lengths are added to a tree in Link notation 
public static String insertLengths(String treeLinkN)
{
 String thisChar = "";
 String newTree = "";
 //length of String   treeLinkN
 int m = treeLinkN.length();
 //TreeLinkN is scanned char by char
 for(int i = 0; i < m; i++ )
  {	 
   thisChar = "" + treeLinkN.charAt(i); 
   newTree = newTree + thisChar;
   if (print1)
    System.out.println(newTree.length() + " " + newTree);
   //Lengths are appended
  if (leaves.contains(thisChar))
   {	
    int k = leaves.indexOf(thisChar);
    double j = lengthsLeaves[k];
    newTree = newTree + ":" + j;  	
   }
 if (internalNodes.contains(thisChar))
  {	
   int k = internalNodes.indexOf(thisChar);
   double j = lengthsInnerLinks[k];
   newTree = newTree + ":" + j;  
  }
 }
 return newTree;	  
}


//Input: string tree with complete information.
//Work: lengthsLeaves[] and lengthsInnerLinks[]
//are read from tree.
private static void decoderLengths(String tree)
{
 if(print2)	
 System.out.println(
	"Input tree at decoverLengths = \n "   + tree  );
 //Delete :
 tree = tree.replaceAll(":", "");
 if(print2)	
  System.out.println(
			"Input tree without : = \n "   + tree  );
 int n = tree.length();
 String thisChar = "";
 String nextChar = "";
 String numbers = ".0123456789";
 boolean leaf = false;
 boolean node = false;
 String Leaf = "";
 String numberLeaf = "";
 String numberNode = "";
 String Node = "";
 int j ;
 double f;

 //All chars in tree are scanned
 for(int i=0; i < n; i++)
 {
  thisChar = "" + tree.charAt(i);
  if (i < n-1) nextChar  = "" + tree.charAt(i+1);
  //Do we have a leaf?
  if (leaves.contains(thisChar)) 
   {
	 leaf = true; 
	 Leaf = "" + thisChar;
   }

  //Do we have an internal node?
  if (internalNodes.contains(thisChar)) 
   {
	 node = true;  
	 Node = "" + thisChar;
   }

  //Do we have a number that belongs in a leaf?
  if ( (leaf) & (numbers.contains(thisChar)))
	{
	 numberLeaf = numberLeaf + thisChar; 
	}

  //Do we have a number that belongs in a leaf?
  if ( (node) & (numbers.contains(thisChar)))
    {
	  numberNode = numberNode + thisChar; 
	 }

  //Must we end accumulation of numbers of a leaf?
  if ( (leaf) & (internalNodes.contains(nextChar)))
	 {
	  leaf = false;
	  j = leaves.indexOf(Leaf);
	  if (print1)
	  System.out.println("Leaf = " + Leaf +
			  " numberLeaf = " + numberLeaf);
	  f = new Double(numberLeaf);
	  lengthsLeaves[j] = (int) Math.floor(f);
	  numberLeaf = "";
	 }

  //Must we end accumulation of numbers of a node?
  if ( (node) & (!(numbers.contains(nextChar))))
	 {
	  node = false;
	  j = internalNodes.indexOf(Node);
	  if (print1)
	  System.out.println("Node = " + Node);
	  f = new Double(numberNode);
	  lengthsInnerLinks[j] = (int) Math.floor(f);
	  numberNode = "";
	 }

 } 	 
}

//Random lengths are added to a tree in Link notation 
public static void generateLengths()
{
 //Lengths with digit values are generated
 for(int i = 0; i < numberOfLeaves; i++)
  {
	 //Zero lengths are not allowed 
	 lengthsLeaves[i] = 1 + r.nextInt(maxLength - 1); 
	 lengthsInnerLinks[i] = 1 +  r.nextInt(maxLength - 1);
  }
} 


//Lengths undergo mutation
private static  void mutateLengths()
{

 //Mutation in links that connect leaves
 for(int k = 0; k < numberOfLeaves; k++)
  {
	 double q = r.nextDouble();		
	 if (q < m) //mutate
			lengthsLeaves[k] = r.nextInt(maxLength);		 	
  }
 //Mutation in links that connect inner nodes
 for(int k = 0; k < numberOfLeaves-2; k++)
  {
	 double q = r.nextDouble();
	 if (q < m) //mutate
			lengthsInnerLinks[k] = r.nextInt(maxLength);		 	
  }
	


}



//Type of  mutation is decided and executed:
//mutation of lengths of links
//else full mutation (topology and lengths of links).
private static String mutateTree(String tree)
{
 //Topology is read from tree
 String topology = deleteLengths(tree);
 //Lengths are read from tree and recorded in vectors
 //lengthsLeaves[] and lengthsInnerLinks[]
 decoderLengths(tree);
 mutateLengths();
 String  mutantTree = "";
 //Mutate lengths else topology ( + new lengths)?
 double s = r.nextDouble();
 if ( s < 0) 
	{
	 if (print1)
	 System.out.println("\n====Mutation of lengths");
	  
	  mutantTree = 	insertLengths(topology);

	  if (print1)
	  {
	  System.out.println("Input tree =  " + tree);
	  System.out.println("Topology =  " + topology);
	  System.out.println("Output tree= " + mutantTree);
	  }
	 }
	else 
	 {
	 if (print1)
	  System.out.println("\n====Mutation of topology");
	 
	  
	//Mutation of topology
	  String mutantTopology = mutation(topology, 
			                           nAdditionalLeaves);
	  //Mutant lengths are reinserted
	  mutantTree = 	insertLengths(mutantTopology);
	 }
	
 return mutantTree;
}



//Individuals are mutated
private static void mutation()
{
 if (print6)
  {
	System.out.println("\n====Mutation \n");
	
	System.out.println("Trees to be mutated");
  }
 for(int i = 0; i < N; i++)
	 {
	  trees[i] = newTrees[i];
	  if (print6)
	  System.out.println(trees[i]);
	 }
 //The old best, i = 0, is conserved intact
 newTrees[0] = trees[0];
 //Other N-1 individuals suffer mutation
 for(int i = 1; i < N; i++)
  newTrees[i] = mutateTree(trees[i]);


if (print6)
	System.out.println("Mutant trees");
	 for(int i = 0; i < N; i++)
	 {
	  trees[i] = newTrees[i];
	  if (print6)
	  System.out.println(trees[i]);
	 }
}

//==========Recombination============


 //Counter of subtopologies
 private static  int counterSubtop;
 //Particular number of subtopologies
 private static  int m1 = 0;



 private static  String[] subtrees1, subtrees2;
 private static  String   subTree1, subTree2,
                     treeRec1, treeRec2;

 //Structure of hashtable : integer + {string1, string2}
 private static Hashtable<Integer, String[]> hashtable =
		 new  Hashtable<Integer, String[]>();

 



//========== Subtrees are listed ============


//Subtrees of  topology are found
private static String[] subtopologies(String topology)
{
 if (print1)
  System.out.println("\nSubtrees are found");
 //Counter of left parenthesis
 int counter = 0;
 //Counter of subtopologies
 counterSubtop = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Length of String   topology
 int m = topology.length();
 String[] subtopologies = new String[m];
 //String topology is scanned char by char
 for(int i = 0; i < m; i++ )
  {	 
   String Sub = "";
   String a = "" + topology.charAt(i);
   int position = i+1;
   if (a.equals(pl))
   {   
    counter++;
    Sub =  Sub + a;
    //The complete subtopology is captured:
    //Parentheses must balanced
   while (counter > 0)
    {   
     String b = "" + topology.charAt(position);
     if (b.equals(pl))   counter++;
     if (b.equals(pr))  counter--;
     Sub =  Sub + b;	 
     position++;
  } 	
 if (print2)
	   System.out.println(Sub);
 subtopologies[counterSubtop] = Sub; 
 counterSubtop++;
 }
}	 
return subtopologies;
}


//Leaves of subtopologies  are found
private static String[]leaves(int m, String[] subtop)
{
 if (print1)  
  System.out.println("Leaves are found");
 String[] myLeaves = new String[m];
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
   else
    {
     myLeaves[i] = "";
     int n = s.length();
     for(int j = 0; j < n; j++)
      {
       String c = "" + subtop[i].charAt(j);
       if (leaves.contains(c))
 	     myLeaves[i] = myLeaves[i] + c;
     }	
   if (print1)
     System.out.println(subtop[i] + " "
                        + myLeaves[i] );
   }	
  }
 //Chars of Leaves are rewritten in alphabet order
 if (print1)
  System.out.println("Leaves are listed orderly" );
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
    else
     {
      s =   myLeaves[i];
      myLeaves[i] = "";
      int n = leaves.length();
   for(int j = 0; j < n; j++)
    {
     String c = "" + leaves.charAt(j);
     if (s.contains(c))
 	   myLeaves[i] = myLeaves[i] + c;
    }	
   if (print1)
   System.out.println(subtop[i] + " " + myLeaves[i] );
  }	
 } 
 return myLeaves;
}



//Inner nodes of subtopologies  are found
private static String[] inner(int m, String[] subtop)
{
 if (print1)
  System.out.println("\nInternal nodes are found");
 String[] myNodes = new String[m];
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
    else
     {
      myNodes[i] = "";
      int n = s.length();
      for(int j = 0; j < n; j++)
       {
        String c = "" + subtop[i].charAt(j);
        if (internalNodes.contains(c))
  	     myNodes[i] = myNodes[i] + c;
       }	
    if (print1)
	System.out.println(subtop[i] + " " + myNodes[i] );
  }	
 }
 //Chars of Leaves are rewritten in alphabet order
 if (print1)
 System.out.println("Inner nodes are listed orderly" );
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
    else
     {
      s =   myNodes[i];
      myNodes[i] = "";
      int n = internalNodes.length();
      for(int j = 0; j < n; j++)
       {
        String c = "" + internalNodes.charAt(j);
        if (s.contains(c))
    	 myNodes[i] = myNodes[i] + c;
       }
  if (print1)
   System.out.println(subtop[i] + " " + myNodes[i] );
 }	
} 
return myNodes;
}


//Subtopologies are compared to decide whether
//or not they have the same set of leaves
//and the same set of inner nodes.
//All pairs that authorize recombination
// are kept in a hastable (=  a table for pairing)
private static boolean comparison(String Leaves1[], 
	                       String Leaves2[],   
	                       String Inner1[], 
	                       String Inner2[]) 
{
 if (print6)
 System.out.println("\nComparison of leaves and nodes");
 Boolean recAllowed = false;
 int m = Leaves1.length;
 int n = Leaves2.length;
 Integer counter = 0;
 for(int i = 1; i < m; i++)
  {
   for(int j = 1; j < n; j++) 
    {
     if ( Leaves1[i].equals(Leaves2[j])) 
      {
       subTree1 = subtrees1[i];	
       subTree2 = subtrees2[j];	
       recAllowed = true;
       if (print6)
        System.out.println(
          "Sets of leaves are EQUAL " + i + 
        "  " + Leaves1[i] + "  " + j + "  " + Leaves2[j] );
        String val[] = {subTree1, subTree2};

        //Recombinant subtrees are kept in a hashtable
        hashtable.put(counter,val);
        counter++;
        if ( Inner1[i].equals(Inner2[j]))   
         {
	      if (print6)
	 	    System.out.println(
	        "Sets of inner nodes are EQUAL " + i +
          "  " + Inner1[i] + "  " + j + "  " + Inner2[j] );
         }
        else 
         {
	       if (print6)
		   System.out.println(
	       "Sets of inner nodes are DIFFERENT " + i +
	       "  " + Inner1[i] + "  " + j + "  " + Inner2[j] );	 
          }
      }
    else 
	  if (print6)
	   System.out.println(
       "Sets of leaves are DIFFERENT " + i +
        "  " + Leaves1[i] + "  " + j + "  " + Leaves2[j] );	
   }
 }	
 return recAllowed;
}


//A pair of subtrees that authorize recombination
//is chosen at random from the hashtable and 
//recombination is executed.
private static  String recombinationTrees(String tree1, 
                                          String tree2)
{
 //Possible pairs of recombinant subtrees are recovered
 //from the hashtable and put into an enumeration<>
 Enumeration<Integer> enumKey = hashtable.keys();
 int counter = 0;
 while(enumKey.hasMoreElements())
  {
   Integer key = enumKey.nextElement();
   String[] val = hashtable.get(key);
   if (print3)
    System.out.println("Hasthtable " + key    + "\n"
		 + val[0]  + "\n"+  val[1]);
   counter++;
   }

 //A pair of recombinant subtrees is chosen at random
 Integer election = r.nextInt(counter);
 if (print6)
  System.out.println("Selected pair = " + election);
 String[] val = hashtable.get(election);
 subTree1 = val[0];
 subTree2 = val[1];
 if (print6)
   System.out.println("Recombinant subtrees \n" +
      subTree1 + "\n" +  subTree2);

 //Recombination is executed
 treeRec1 = tree1.replace(subTree1, subTree2);
 treeRec2 = tree2.replace(subTree2, subTree1);
 if (print6)
  System.out.println("\nRecombination is allowed:"
	 + "\nOriginal trees \n"
	 + tree1 + "\n" + tree2
	 + "\nSwapping subtrees \n" 
	 +  subTree1 + "\n" + subTree2
	 + "\nRecombinant trees \n"
	 + treeRec1 + "\n" + treeRec2 );
 return treeRec1;
}


//Recombination of tree1 and tree2 is tried out
public static String recombinationMaster(String tree1, 
		                              String tree2) 
{   
 //First topology  
 String topo1 = deleteLengths(tree1);
 if (print6)
  {
   System.out.println(topo1 + " -> " 
       + linkToNewick(topo1));
   System.out.println("\n Tree1 = " + tree1 +
			 "\n Topo1 = " + topo1);
  }


 //Length of String   topology
 m1 = topo1.length();

 subtrees1 = new String[m1];
 subtrees1 = subtopologies(topo1);
 m1 = counterSubtop;
 //Subtopologies are found
 String[] subtopologies1 = new String[m1];
 for(int i = 0; i < m1; i++)
  subtopologies1[i] =  simplify(subtrees1[i]);
 if (print6)
 System.out.println();
 String[] Leaves1 = new String[m1];
 Leaves1 = leaves(m1,subtopologies1);
 String[] Inner1 = new String[m1];
 Inner1 = inner(m1,subtrees1);
 //Chars of Leaves are rewritten in alphabet order
 if (print5)
  System.out.println("\nSubtrees, subtopologies, Leaves, " +
	             "Inner nodes" );
 for(int i = 0; i < m1; i++)
  {
   if (subtopologies1[i] == null);
   else
    {   
	 if (print5)  
        System.out.println(i 
		   + "   " + subtrees1[i]
				   + "   " + subtopologies1[i]
		   + "   " +  Leaves1[i] + "   " + Inner1[i]);
    }
  }

 if (print6)
  System.out.println();

 //Second topology 
 String topo2 = deleteLengths(tree2);
 if (print6)
  {
   System.out.println(topo1 + " -> " 
   + linkToNewick(topo2));

	 System.out.println("\n Tree2 = " + tree2 +
			 "\n Topo2 = " + topo2);	
  }
 //Length of String   topology
 m1 = topo2.length();

 subtrees2 = new String[m1];
 subtrees2 = subtopologies(topo2);
 m1 = counterSubtop;
 //Subtopologies are found
 String[] subtopologies2 = new String[m1];
 for(int i = 0; i < m1; i++)
  subtopologies2[i] =  simplify(subtrees2[i]);


 if (print6)
  System.out.println();
 String[] Leaves2 = new String[m1];
 Leaves2 = leaves(m1,subtopologies2);
 String[] Inner2 = new String[m1];
 Inner2 = inner(m1,subtrees2);
 //Chars of Leaves are rewritten in alphabet order
 if (print5)
  System.out.println("\nSubtopologies, Leaves, " +
	             "Inner nodes" );
 for(int i = 0; i < m1; i++)
  {
   if (subtopologies2[i] == null);
   else
    {   
	 if (print5)
      System.out.println(i 
		   + "   " + subtrees2[i]
				   + "   " + subtopologies2[i]
		   + "   " +  Leaves2[i] + "   " + Inner2[i]);
    }
  }   
 boolean recAllowed = 
	   comparison(Leaves1,Leaves2, Inner1,Inner2); 
 String recombinant = "";
 if (recAllowed)
   recombinant = recombinationTrees(tree1,tree2);
 else 
  {
   recombinant = tree1;
	
   if (print6)
	  System.out.println("Recombination is not allowed");
  }
 return recombinant;
}


//Recombination methods may produce an output
//with various and different assignment of lengths to 
//inner node a. One is chosen.
public static void correction() 
{ 
	String topo;
 for(int i = 1; i < N; i++)
  {
	decoderLengths(newTrees[i]);
	topo = deleteLengths(newTrees[i]);
	newTrees[i] = insertLengths(topo);
  }
}

//Each one of the N individuals  are tried out
//for recombination: a partner is chosen at random
//and if recombination is allowed, it is executed.
//Otherwise, it is cloned.
public static void recombination() 
{ 
 if (print6)	
 System.out.println( "\n====RECOMBINATION \n");
 //Best individual is excluded from recombination
  newTrees[0] = trees[0];

 for(int i = 1; i < N; i++)
  {
   //A partner for i is chosen	 
   int j = i;
   String t1 = "";
   String t2 = "";
   while (i == j)
	 j =   r.nextInt(N);
   if (print6)	
	System.out.println( "\n==RECOMBINATION OF INDIVIDUALS "
			+ i + " and " + j);   
   t1 = trees[i];
   t2 = trees[j];   
   newTrees[i] = recombinationMaster(t1,t2);
   correction();
  }
 if (print6)
 System.out.println("Recombination output");
 for(int i = 0; i < N; i++)
 {
  trees[i] = newTrees[i];
  if (print6)
  System.out.println(trees[i]);
 }
}



//Evolution of trees involving
//recombination and mutation of topologies,
//in within and between.
private static void evolution()
{
 for(int i = 1; i < numberOfGenerations; i++)
 {
	generation = i;
  if ( (i % lapse) == 0) 
		  print7 = true;
  else print7 = false;
  
  if (print7)
  System.out.println( "\n===GENERATION  = " + i);
  selectiveCloning();
  mutation();
  recombination();
  assessFitnesses();
  scaleFitnesses();
 
 }
}





//A tree topology is generated at random.
public static String randomTopology()
 {
  if (print2)
  System.out.println("Parent tree = \n" + seed
 		 + " -> "    + simplify(seed)); 
  String tree = seed;
  int limit = numberOfLeaves-3; //less than 11.
  for(int i = 0; i < limit; i++ ) 
	  tree = generateChild(tree);
  System.out.println("Target topology = \n" + tree);
  return tree;
 }
 
 
//======================
  

  

//The best fitting tree for chosen topology is found
public static int findTheBest(String tree)
{
 String selectedTopology = simplify(tree);	
 String thisTopology = "";
 String thisTree = "";
 int bestPosition = -1;

for(int i = 0; i < indexOfTopologies; i++)
 {
  thisTree = bestTreeOfTopology[i];
  thisTopology = simplify(thisTree);
  if (selectedTopology.equals(thisTopology))
  {
   bestPosition = i;
   return bestPosition;	    
  }
 }
return  bestPosition;
}


  /*
  //Lengths with real values are generated
  double lengthsLeaves[]  = new double[l];
  //Last two entries are redundant
  double lengthsInnerLinks[]  = new double[l];  
  for(int i = 0; i < numberOfLeaves; i++)
   {
	 lengthsLeaves[i] = r.nextDouble(); 
	 lengthsInnerLinks[i] =  r.nextDouble();
   }
  */  

  
  
  //Data about the tree is printed to the console
  public void report()
  {
	System.out.println("The root has " + 
      treePanel.rootNode.getChildCount() + " children");
	System.out.println("The tree's depth is " + 
      treePanel.rootNode.getDepth());
	System.out.println("The tree has " + 
      treePanel.rootNode.getLeafCount() + " leaves");
	 System.out.println("'Root' is really a root? " + 
      treePanel.rootNode.isRoot());
	 System.out.println("Root's userObject: " + 
      treePanel.rootNode.toString());	 
  }
  
  
  /**
   * Create the GUI and show it. For thread safety, 
   * this method should be
   * invoked from the event-dispatching thread.
   */
private static void createAndShowGUI()
 {
  // Create and set up the window.
  JFrame frame = 
		new JFrame("Tree of tree topologies.");
  frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);    
  // Create and set up the content pane.
  
  // content panes must be opaque
  newContentPane.setOpaque(true); 
  frame.setContentPane(newContentPane);

  // Display the window.
  frame.pack();
  frame.setVisible(true);
 }









//========= Translation in within formats ==========
  
  

//bTree in (modified) Newick's notation 
//is used to generate Record[]
//with addresses and contents.
public static void makeInput(String bTree) 
{
 if (print)
 System.out.println("bTree at makeInput ////// " + bTree);
 int counter = 0;
 //length of String   tree
 int l = bTree.length();
 String thisChar = "";
 String address = "";
 String binNumbers = "01";
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + bTree.charAt(i);
	if (binNumbers.contains(thisChar)) 
		   address = address + thisChar;
	if (leaves.contains(thisChar)) 
	{
	 Record[counter] = address;
     counter++;
	 Record[counter] = thisChar;
	
	counter++;
	address = "";
   }
  }
 
 if (print)
for(int i = 0; i < counter; i = i+2 )
	  System.out.println(Record[i] +
			 " bbbb " + Record[i+1]);

lastUsedPosition = counter;
if (print)
	System.out.println("lastUsedPosition ==== " 
        + lastUsedPosition); 
}



//Outputs a word in reversed order.
private static String reverser(String word)
{
 return new StringBuilder(word).reverse().toString();	
}


private static String cleanTree(String tree)
{
 if (print2)
 System.out.println("Input tree at cleanTree = \n "
        + tree);
 String treeClean = "";	
 String numeric = ".:0123456789";
 String plusMinus = "+-";
 String thisChar = "";
  boolean leaf = false;
 //Numerics are washed
 //length of String   tree
 int l = tree.length();
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + tree.charAt(i);
   if (numeric.contains(thisChar));
    else treeClean = treeClean + thisChar;	  
  }
 if (print2)
 System.out.println("Tree without numerics = \n"
		+ treeClean);
 //+/- are deleted from inner nodes that accompany leaves
 l = treeClean.length();
 String treeOut = "";
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
 {	 
  thisChar = "" + treeClean.charAt(i);
  if (leaves.contains(thisChar))
	  leaf  = true;
  if ((leaf) &  (plusMinus.contains(thisChar)));
	else treeOut = treeOut + thisChar;	  
  if (internalNodes.contains(thisChar))
	  leaf  = false;
}

 if (print2)
 System.out.println("Tree at output of cleanTree = \n"
			+ treeOut);
return treeOut;

}




//The path from each leaf to node a is found.
//The input contains branch's indicator, say:
//((++Aa,+-Ba),(-+Cb,((--++Dd,--+-Fd)--c,---Ec)-b)+a)
//Output for leaf A: Aa+.
//Output for leaf C: Cba-
//Output for leaf D: Ddcba-
private static void findPaths(String tree)
{
 if (print2)	
 System.out.println("Input tree at findPaths = \n" + tree); 
 //length of String   tree
 int l = tree.length();
 String plusMinus = "+-";
 String thisChar;
 String nextChar;
 String path = "";
 //Branch indicative with respect to root, + else -
 String branch;
 String address ;
 //The number of links of a path to root
 int level = 0;
 String treeReversed = reverser(tree);
 if (print2)
 System.out.println("Reversed tree = " + treeReversed);
 
//Each leaf has a path to a	 
for(int k= 0; k < numberOfLeaves; k++)	 
{
 String leaf = "" + leaves.charAt(k); 
 //Candidate nodes to be part of the path to node a
 //must have  sign +/- as prefix.
 //Or as suffix in the reversed tree.
 int position = treeReversed.indexOf(leaf);
 if (position > -1) // if leaf is in tree
 {
  if (print2)
  System.out.println("\nLeaf = " + leaf ); 
  address = "";
  boolean go = true;
  //branchIndicator and level are found
  for(int i = position+1 ; ((i < l) & (go)); i++ )
  {	 
   thisChar = "" + treeReversed.charAt(i);  
   if   (plusMinus.contains(thisChar)  )
	address = thisChar + address;
   else go = false;   
  }  
  level = address.length();
  if (print2)
  System.out.println("LeafAddressLevel  = " + leaf 
		                     + address + level); 
  branch = "" + address.charAt(0);
 
  
  //Path from leaf to node a is found in (not reversed) tree 
  position = tree.indexOf(leaf);
  thisChar = "" + tree.charAt(position); 
  nextChar = "" + tree.charAt(position + 1);
  //Output
  path = "" + leaf + nextChar;
  go = true;
  if  (nextChar.contentEquals("a") ) //leaf is neighbor of a
   {
	  path = path + branch;
	  if (print)
	  System.out.println("Path short = " + path);
   }
  else	 //level > 1
  {	 	  
	  
  //Tree is scanned char by char to get path
  String iChar = "";
  for(int i = position + 2; ((i < l) & (go)); i++ )
  {	 
   iChar = "" + tree.charAt(i);
   
  
   int newLevel = 0;
   if (internalNodes.contains(iChar))//InternalNodes
	{	   
     if   (iChar.contentEquals("a")  )
      {
	   path  = path +  iChar;
	   go = false; //end
      }
     else //Internal node different than a
     {
      if (print)	 
      System.out.println("Internal node not a) = " + iChar );
      int rP = l -i-1;	//index of iChar in treeReversed       
      String  nextNewChar = "" + treeReversed.charAt(rP+1);
      
      if ( plusMinus.contains(nextNewChar )) //candidate node
       {  
    	  if (print)  
    	System.out.println("next candidate = " + iChar); 
        String addressNew = "";
        
        //level of candidate node iChar  is inquired
        String thatChar = "";
        boolean go2 = true;
        for(int j = rP+1 ; ((j < l) & (go2)); j++ )
         {	 
    	  thatChar = "" + treeReversed.charAt(j);  
    	  if   (plusMinus.contains(thatChar)  )
          addressNew = thatChar + addressNew;
    	  else go2 = false;
         }  
        if (print)
        System.out.println("AddressNew " +
                                    iChar + addressNew);
       newLevel = addressNew.length();
       //A path links a node with another in a lower level
       if  ( (newLevel < level) &
    		   (!(nextChar.contentEquals(iChar))))
        {
    	  path = path + iChar;
    	  level = newLevel;
    	  if (print)
    	  System.out.println("Path newLevel= " + path);
        }
      } //End of if candidate node    	 	 
     }//End of else Internal node different than a 
	}//End of if InternalNodes 
   }//End of Tree is scanned char by char
  
  path = path + branch;
  
 }//End of  else level > 1    
 }//End of if leaf is in tree
 if (print2)
 System.out.println("Path completed  = " + path);
 Paths[k] = path;
}//End of each leaf
}



//Paths from leaves to node a are found
//Reverse paths are also calculated.
//Input tree in quasibinary notation 
//with addresses and lengths
private static void pathFinder(String tree)	
{
 //Useless information is deleted	
 String tree2 = cleanTree(tree);
 //Find the path to node a from each leaf.
 //Add an indicator of the branch
 //with respect to the root: up(+) else down(-)
  findPaths(tree2);  	
}

//Input: a long raw string with path from leaf to node a
//plus path from node a to leaf.
//Editing is done to get true paths.
private static String grammar (String path)
{
 int n = path.length();
 String thisChar = "";
 String nextChar = "";
 String oldPath = path;
 String plusMinus = "a+-a";
 String minusPlus = "a-+a";
 String two = "";
 int N = path.length();
//Rule 1: two equal symbols produce nothing
 for(int j=1; j < N; j++)
  {
   for(int i=1; i < n-1;)
   {	 
    thisChar = "" + path.charAt(i);
    nextChar = "" + path.charAt(i+1);
    two = "" + thisChar + nextChar;
    if (thisChar.contentEquals(nextChar)) 
	  {
	   i = i + 2; 
	   path = path.replace(two, "");
	  }
    else 
	  {	   
	   i = i+1;
	  } 
  n = path.length();
   } 
  }
 
 //Rule 2: a+-a -> a
 
 if  (path.contains(plusMinus))		
      path = path.replace(plusMinus, "a");
 if   (path.contains(minusPlus))
     path = path.replace(minusPlus, "a");
 if (print2)   
 System.out.println("input = " + oldPath + 
		                    " output = " + path);
 return path;	
}

//The distance matrix among leaves is calculated
private static double[][] matrixBuilder()
{
	double M[][] = 
	new double[numberOfLeaves][numberOfLeaves];
 //Path to node a from leaf
 String pathLeafToa = "";
 //Path to leaf from node a
 String pathaToLeaf = "";
 //Path from leaf to lead
 String pathLeafiToLeafj = "";
 String innerNode = "";
 double distance;
 for(int i = 0; i < numberOfLeaves; i++)
  for(int j = 0; j < numberOfLeaves; j++)
  {
   pathLeafToa = Paths[i];
   pathaToLeaf = reverser( Paths[j] );
   //Raw paths
   pathLeafiToLeafj = pathLeafToa + pathaToLeaf;
   //An editing grammar to get true paths
   pathLeafiToLeafj = grammar( pathLeafiToLeafj );
   if (i == j) M[i][j] = 0;
   else
    {	 
	 String leaf = "" + pathLeafiToLeafj.charAt(0);
	 int p = leaves.indexOf(leaf);	 
	 distance = lengthsLeaves[p];
	 int n = pathLeafiToLeafj.length();
	 for(int k = 1; k < n-1; k++)
	 {	 
	  innerNode =  "" + pathLeafiToLeafj.charAt(k);
	  p = internalNodes.indexOf(innerNode);	 
	  distance = distance + lengthsInnerLinks[p];
	 }
	 leaf =  "" + pathLeafiToLeafj.charAt(n-1);
	 p = leaves.indexOf(leaf);	 
	 distance = distance + lengthsLeaves[p];
	  M[i][j] = distance;
    }
  }
 return  M;
}



//Translation from Newick's to quasi-binary notation:
//addresses are attached to tree in Newick's notation
//Inner nodes are conserved. Lengths are conserved.
private static String toQuasiBinaryInner(String tree) 
{
 if (print2)
  System.out.println("Input tree at toQuasiBinaryInner " +
		                         "= \n" + tree); 	
 //Output
 String treeToBi = "";	
 String numeric = ".:0123456789";
 //The binary address of leaf
 String address = "";
 //The name or content of the tree
 String content = "";
 String thisChar = "";
 String nextChar = "";
 String previousChar = "";
 //length of String   tree
 int l = tree.length();
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + tree.charAt(i);
   if ( i < l-1) nextChar = "" + tree.charAt(i+1);
    else nextChar = "end";
   if ( i > 0) previousChar = "" + tree.charAt(i-1);
    else previousChar = "start";


   //This char = (
   if (thisChar.contentEquals("(")) 
    {
     content = content + thisChar;
     address = address + '+';
     if (print)
	 System.out.println("Address = " + address); 		
    }
	   
	   
 //If this char is a leaf 
 if (leaves.contains(thisChar))
  {
   // System.out.println("leaf"); 
   treeToBi =  treeToBi + content + address + thisChar;
   content = "";
  }

 if (numeric.contains(thisChar))
	 treeToBi =  treeToBi + thisChar;
	 
 //If this char is an inner node
 if (internalNodes.contains(thisChar))
  {
   int m = address.length();  
			//delete last bit from the right
   if (m>0)	address = address.substring(0, m-1) ;
   if (address.contentEquals("")) address = "+";
   treeToBi =  treeToBi +	address + thisChar;	 		
   
  }
	 
	 
 //This char = , and goes after an inner node
 if  (thisChar.contentEquals(",")) 
  {
   //System.out.println
	 //("comma : "  + previousChar + thisChar ); 
   treeToBi =  treeToBi +	thisChar ;
   if (print)
	 System.out.println("Address at ,1   = " + address); 
		   
   if    (numeric.contains(previousChar))
		  address = address + '-';//Add a - to the right
   if (print)
	 System.out.println("Address at ,2   = " + address);
  } 
	   
	   
 if (thisChar.contentEquals(")")) 
  {
	treeToBi =  treeToBi +	thisChar ;
    //Diphthonge   ),
	if   (nextChar.contentEquals(","))
	{ 
     int m = address.length(); 
	 //delete last bit from the right  
	 if (m >0)   address = address.substring(0, m-1) + "-";
   }
  }
	  		
}//End of for(int i...
return treeToBi;	
}


//bTree  (in links with lengths notation) 
//is used to generate Record[]
//with addresses and contents.
//Names of inner nodes are visible
public static void makeInputInner(String bTree) 
{
 int counter = 0;
 //length of String   tree
 int l = bTree.length();
 String thisChar = "";
 String address = "";
 String binNumbers = "+-";
 String numeric = ".:0123456789";
 //String tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + bTree.charAt(i);
	if (binNumbers.contains(thisChar)) 
		   address = address + thisChar;
	if ( (leaves.contains(thisChar)) ||
			(internalNodes.contains(thisChar)))
	{
	 Record[counter] = address;
     counter++;
	 Record[counter] = thisChar;
	
	 counter++;
	 address = "";
    }
	
	if (numeric.contains(thisChar))
	 {
		Record[counter-1] = Record[counter-1] + thisChar;
	 }
	
  }


 if (print)
  for(int i = 0; i < counter; i = i+2 )
	  System.out.println(Record[i] +
			 " bbbb " + Record[i+1]);

 lastUsedPosition = counter;
 if (print)
	System.out.println("lastUsedPosition ==== " 
      + lastUsedPosition); 
}//End of method


//Prints  a matrix and its title
private static void printMatrix(double Matrix[][], 
		                        String title)
{
 System.out.println( "\n" + title );
 int m = Matrix.length;
 System.out.print(  "\t");
 for ( int j = 0; j < m; j++ )
  System.out.print( leaves.charAt(j)+ "\t");		
 System.out.println();
 for ( int i = 0; i < m; i++ )
  {
   System.out.print(leaves.charAt(i ) + "\t");	
   for ( int j = 0; j < m; j++ )
	 System.out.print( Matrix[ i ][j]+ "\t");
   System.out.println();
  }
}


//Tree is translated 
//from link notation to binary.
//Output in Record[]: even entries= addresses, odd = names.
//Associated matrix is also calculated
public static double [][] associatedMatrix(String tree) 
{
 /*	
 System.out.println("Tree  at translate = " + tree);
 //String NwTree = simplify(tree);
 String bTree = toBinary(tree);
 //A vector with addresses and names is composed
 makeInput(bTree);
 */
//String NwTree = simplify(tree);
 String bTree = toQuasiBinaryInner(tree);	
 
 if (print2)
 {	 
 System.out.println("Input tree  at translate = "    + tree);
  System.out.println("Selected tree in " +
	 	"pseudo-Newick's notation = \n" + simplify(tree));
  System.out.println("Tree in quasi-binary notation: \n"
		               + bTree);
 }
 
 //A vector with addresses and other with names are composed
 makeInputInner(bTree);	
 //Paths from leaves to node a are found.
 //Reverse paths are also listed.
 pathFinder(bTree);	
 //The distance matrix among leaves is calculated
 double M[][] = matrixBuilder();
 return M;
}
 




//The absolute distance among two matrices is calculated.
//Only triangular part of matrices are compared. 
private static double absDistance(int n, 
		               double[][] M1, double[][] M2)
{
double abs = 0;
for(int i = 0; i < n; i++)
for(int j = i+1; j < n; j++)
 abs = abs + Math.abs(M1[i][j]-M2[i][j]);  
return abs;
}


//Input : tree in link notation.
//Output:  
//distance matrix associated to the leaves of the tree.
private static double[][] distanceMatrix(String tree)
{
	String s = toQuasiBinaryInner(tree) ;	
	if (print4)
	 System.out.println("\n Tree = " + s);	
	//Paths from leaves to node a are found.
	//Reverse paths are also listed.
	pathFinder(s);	
	//The distance matrix among leaves is calculated
	double M[][] = matrixBuilder();	
	return M;
}


















 
  

//=========================================================
//===================Inner service class===================
//=========================================================

  

class DynamicTree15 extends JPanel 
{

 private static final long serialVersionUID = 1L;
	
 protected DefaultMutableTreeNode rootNode;
 protected DefaultTreeModel treeModel;
 protected JTree tree;
 //Next leaf to be grafted	
 char nextLeaf = leaves.charAt(3);
 //Next internal node to be inserted
 char nextIntNode = internalNodes.charAt(1);
 //counter of generations
 int counterGen = 0;

 
 //Constructor
 public DynamicTree15() 
 {
    super(new GridLayout(1, 0));
    
    rootNode = new DefaultMutableTreeNode(seed);
    treeModel = new DefaultTreeModel(rootNode);

    tree = new JTree(treeModel);
    tree.setEditable(true);
    tree.getSelectionModel().setSelectionMode(
        TreeSelectionModel.SINGLE_TREE_SELECTION);
    tree.setShowsRootHandles(true);

    JScrollPane scrollPane = new JScrollPane(tree);
    add(scrollPane);
  }

 
  /** Selected tree is unveiled */
  public String unveilTree() 
  {
    DefaultMutableTreeNode selectedNode = null;
    TreePath selectedPath = tree.getSelectionPath();

    if (selectedPath == null) {
    	selectedNode = rootNode;
    } else {
    	selectedNode = (DefaultMutableTreeNode) 
    		        (selectedPath.getLastPathComponent());
    }
    
    String tree =  selectedNode.toString();
    if (print)
    System.out.println("\nSelected tree at unveilTree() = "
                         + tree);
    return tree;
  }

 
  //Add a node
  public DefaultMutableTreeNode 
               addObject(DefaultMutableTreeNode seed,
      Object child, boolean shouldBeVisible) {
    DefaultMutableTreeNode childNode = 
    		           new DefaultMutableTreeNode(child);

    if (seed == null) {
      seed = rootNode;
    }

    // It is key to invoke this on the TreeModel, 
    //and NOT DefaultMutableTreeNode
    treeModel.insertNodeInto
              (childNode, seed, seed.getChildCount());

    // Make sure the user can see the lovely new node.
    if (shouldBeVisible) {
      tree.scrollPathToVisible
                     (new TreePath(childNode.getPath()));
    }
    return childNode;
  }



		
	
  
  //Listener
  class MyTreeModelListener implements TreeModelListener
  {
    public void treeNodesChanged(TreeModelEvent e) 
    {
      DefaultMutableTreeNode node;
      node = (DefaultMutableTreeNode) 
    		    (e.getTreePath().getLastPathComponent());

      /*
       * If the event lists children, 
       * then the changed node is the child of the
       * node we've already gotten. 
       * Otherwise, the changed node and the
       * specified node are the same.
       */

      int index = e.getChildIndices()[0];
      node = (DefaultMutableTreeNode) 
    		            (node.getChildAt(index));

      System.out.println(
    		   "The user has finished editing the node.");
      System.out.println("New value: " + 
    		             node.getUserObject());
    }

    public void treeNodesInserted(TreeModelEvent e) 
    {
    }

    public void treeNodesRemoved(TreeModelEvent e) 
    {
    }

    public void treeStructureChanged(TreeModelEvent e) 
    {
    }
  }
}//End of DynamicTree15



//===========================================
//==Inner class  for drawing selected tree ==
//===========================================


//This program translates a string that 
//denotes a tree in Newick's notation
//to our binary notation
//(assign a zero to the left, a one to the right).

//A generator of random trees in Newick's notation
//is included.

//Resultant trees are drawn.


	


class DrawBiTree12 extends JPanel
                      
{
 private static final long serialVersionUID = 1L;
 private JPanel drawingPane32;
 JScrollPane scroller;


 

 //=======================================



 //Vertical  size of the panel 
 int verticalSize;
 int horizontalSize;
//Scale must be a power of 2
 int scale = 1;
	
 //Symbolic horizontal length of each stroke
 int horizontalLength = 30;

 int verticalLength = 10;




 String  NwTree = "";



 boolean print = false;



 //This is our encoding for a tree in our binary notation:
 //We use a vector of strings:
 //Even entries are the paths in the tree  of the nodes,
 //odd ones are the strings, names or contents.



 //Y-coordinates of points where strings are drawn
 int[] yCoordinate = new int[1000];



 //Set up the instructions.
 JLabel instruct1; 
 JLabel instruct2;

public DrawBiTree12() 
{
 super(new BorderLayout());

     
 JPanel instructionPanel = new JPanel(new GridLayout(0,1));
       

 //Set up the instructions.
 instruct1 = new JLabel(title1);
 instruct2 = new JLabel(title2);

 instructionPanel.setFocusable(true);
 instructionPanel.add(instruct1);
 instructionPanel.add(instruct2);

 //Set up the drawing area.
 drawingPane32 = new drawingPane32();
 drawingPane32.setBackground(Color.white);

 //Put the drawing area in a scroll pane.
 JScrollPane scroller = new JScrollPane(drawingPane32);
 scroller.setPreferredSize(new Dimension(500,900));
 scroller.revalidate();

 //Lay out this demo.
 add(instructionPanel, BorderLayout.PAGE_END);
 add(scroller, BorderLayout.CENTER);
}


//=========================================
//============= Painting class ============
//=========================================


/** The component inside the scroll pane. */
 class drawingPane32 extends JPanel 
{
 private static final long serialVersionUID = 1L;
	   
 //Initialization of coordinates
 int centerX = 1;
 int centerY = 1;
 //Coordinates of the origin
 int x00;
 int y00;
 //Target coordinates
 int x11 ;
 int y11;
 //Horizontal and basic vertical strokes
 int H;
 int V;
 //Addresses are interpreted as paths	
 String path = "";
 char direction;




//Calculate the size of a box to hold the tree
//Scrolling controls are managed
private void sizeOfTree(Graphics2D g2d)
{
 //Calculate the dimension of the strings
 //All strings have equal size.

 //Get   font in use
 Font font = g2d.getFont();
 //get metrics from the graphics
 FontMetrics metrics = g2d.getFontMetrics(font);
 //get the advance of my text in this font
 //and render context
 int lengthText = metrics.stringWidth("Aa") + 2;
 //get the height of a line of text in this
 //font and render context (some padding is included)
 int hgt = metrics.getHeight()  + 8;
 if (print)
  System.out.println("height of letters = " + hgt);

 //Horizontal and basic vertical strokes
 H = scale * horizontalLength;
 V = scale * verticalLength;

 horizontalSize = lengthText +
              (int) (H *  numberOfLeaves);
 verticalSize = 
		 (int) (hgt *   numberOfLeaves);


 //area = size of the graphic (width, height)
 Dimension area = 
		 new Dimension(horizontalSize,12*verticalSize);
 //area is set as the preferred size of the graphic
 drawingPane32.setPreferredSize(area); 
 //The graphic demands a space for itself:
 //if the alloted space is not enough, 
 //turn on scrolling bars.
 drawingPane32.revalidate();
 verticalLength = 2*verticalSize;
 //Coordinates of the origin
 x00 =    scale*centerX;
 y00 =   scale* centerY + verticalSize  - 600;
 //Target coordinates
 x11 =  x00 + (int) ( scale*horizontalLength);
 y11 =  y00;

}



//Tree is drawn
private void DrawTree3(Graphics2D g2d)
{	
//The tree is drawn:
//A path from the origin of the tree to
//each leaf is drawn
 if (print)
 System.out.println("LastUsedPositoon at DrawTree3=  " 
                              + lastUsedPosition);
 for(int i = 0; i < lastUsedPosition; i = i+2)
  {
	//Addresses will be used as paths to  be drawn
	path = Record[i]; 
	if (print)
	System.out.println( i + " Path = " + path);
	//Reinitialization
	Dimension center = new Dimension(x11,y11);      
	V = scale * verticalLength;
	//Path is interpreted
	for(int j = 0; j < path.length(); j++)
	 {
	  direction = path.charAt(j);
	  int x0 =  center.width;
	  int y0 =  center.height;
	  int x1 =  x0;
	  int y1;
	  if (direction == '+') //go up
	   {      		
	    y1 =  y0 + V;       		
	    center.height = center.height + V;
	   }	
	   else // if direction == '-'  =  go down	 
	   {  
	     y1 =  y0 - V;        	
	     center.height = center.height - V;
	    } //End of if + else      
	    //Draw vertical stroke
	    g2d.drawLine(x0, verticalSize -y0, x1,
	    				verticalSize- y1);
	    //Draw horizontal stroke
	    g2d.drawLine(x1, verticalSize - y1, 
	    		   x1 + H,  verticalSize - y1);
	    //Move center       	
	    center.width = center.width + H ;   
	    //Vertical stroke is divided in two
	    V = V/2;    	
	    }//End of for j        	            
	 }  //End of for: numberOfLeaves    	  
}



//Strings are drawn
private void drawStrings(Graphics2D g2d)
{
 int counter = 0;	
 //Seed is not  drawn: we lack one
 //g2d.drawString(seed,x00,verticalSize - y00);
 yCoordinate[counter] = verticalSize - y00;
 //counter++;
 //Strings are drawn:   
 for(int i = 0; i < lastUsedPosition; i = i+2)
 {
  //Addresses will be used as paths to the places
  //where strings must be drawn
  path = Record[i];
  String s = Record[i+1]; 
  //Reinitialization
  Dimension center = new Dimension(x00,y00);
  Dimension target = center;
  V = scale * verticalLength;
  //Path is interpreted
  for(int j = 0; j < path.length(); j++)
	{
	 direction = path.charAt(j);
	 if (direction == '+') //go up
	 target.height = target.height + V;    	 
	 else // if direction == '-'  =  go down	 
	 target.height = target.height - V;
	 target.width = target.width + H ;
	 V = V/2;    	
	}//End of for j
	//String is drawn
	g2d.drawString(s, target.width + 1, 
	                      verticalSize - target.height-1); 
	//y-coordinates are recorded
	yCoordinate[counter] = verticalSize - target.height-4;
	 counter++;
  }    
} 	


//Basic routine of the drawing procedure
private void drawBasic(Graphics2D g2d)
{
 g2d.setColor(Color.red);
 //Stem is drawn: our trees is artificially rooted.  
 g2d.drawLine(x00, verticalSize - y00, 
	  		   x11,  verticalSize - y11);
 sizeOfTree(g2d);
 DrawTree3(g2d);
 drawStrings(g2d); 
}

//Graphical  master	: it is called automatically 
//at starting and by repaint()
protected void paintComponent(Graphics g) 
{
 super.paintComponent(g);
 Graphics2D g2d = (Graphics2D)g;
 //Font to draw strings
 Font font = new Font("Arial", Font.PLAIN, 12);  
 g2d.setFont(font);
 g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
		  RenderingHints.VALUE_ANTIALIAS_ON);  
 drawBasic(g2d);   
}


}//End of double inner class drawingPane32



/**
* Create the GUI and show it.  For thread safety,
* this method should be invoked from the
* event-dispatching thread.
*/
 void createAndShowGUI() 
{
 //Create and set up the window.
 JFrame frame = 
   new JFrame(titleFrame);
 frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);

 //Create and set up the content pane.
 //Constructor of the main application is used.
 JComponent newContentPane = new DrawBiTree12();
 //content panes must be opaque
 newContentPane.setOpaque(true); 
 frame.setContentPane(newContentPane);

 //Display the window.
 frame.pack();
 frame.setVisible(true);
}

}//End of  inner class DrawBiTree12

}//End of main class K38 EvoFull