//Program K65 EvolFull10
/*
 * We implement here 
 * a full fledged evolutionary process
 * to propose an approximate solution to the problem of
 * fitting a tree to a distance matrix.
 * 
 * This code incorporates into
 * program K57 EvoFull6 the following operations:  
 * swapping of leaves,
 * restructuring of complements of subtopologies
 *  and correlated tiny double mutations of the form 
 * (+/-1, +/-1).
 *
 * The program runs a certain number of generations,
 * reports the tree that best approximates the  solution, 
 * draws the  best approximation tree together with
 * the target tree  and halts.
 * 
 * We generate a random tree topology 
 * and attach random integer lengths to links.
 * In that way we get the target tree, T. 
 * Next, we find the associated distance matrix D  to T.
 * Our problem is:
 * 
 * Find tree t such that the absolute distance between
 * matrix D and the associated matrix to t is a minimum.
 * Absolute distance between M and N = sum |Mij-Nij|.
 * This distance defines the FITNESS of each tree.
 * It decreases as good performance augments.
 * The best possible fitness is zero.
 * 
 * We run the next test:
 * If we begin with a tree T, compute D, and then t,
 * we expect to return with the initial tree: t = T.
 *
 *
 * To associate a distance matrix to a tree, 
 * we calculate the distance in within 
 * each pair of nodes (leaves + inner nodes).
 * To that aim, we begin with the first node and 
 * walk over the tree until the second node
 * and add the lengths of all traversed links.
 * The resulting number is the  distance.
 *  
 *  
 * Formats = Link and Newick's notations. And variants. 
 * 
 * An automatic halting option has been added to be applied
 * if perfect matching is found.
 * The program accepts for input 
 * a distance matrix with integer values.    
 * A test for being a distance matrix is run.
 * 
 * To find the tree that best fit D, 
 * we  use evolution as follows:
 * 
 * 1) Initialization:
 * 		We generate N topologies  on the same
 * 		set of leaves. By adding   lengths to links,
 * 		every topology gets represented by a tree.
 *		The fitness of each tree
 *		is calculated and the best is selected.
 * 		If the fitness of the best is zero, 
 * 		the program halts and the tree is reported.
 * 
 * 2) Evolution is composed of rounds of
 *   	 cloning according to fitness, 
 *   	 random mutation and recombination.
 *       After each generation, fitness is checked 
 *   	 for accomplishment. 
 * 
 * 
 * Mutation is Lamarckian: 
 * it always produces viable mutants. 
 * Topologies as well as links can mutate.
 * 
 * A mutation results from deleting 
 * a given number of leaves
 * chosen at random
 * to reinsert them anew in 
 * other places chosen at random.
 * This creates a mutant topology.  
 * To mutate length of links:
 * Lengths are read from tree and 
 * mutated separately.
 * Then, they are reinserted.
 * 
 * 
 * Recombination is Lamarckian: it is authorized only
 * when it produces viable recombinants.
 * All their subtopologies are calculated 
 * and packed in arrays.
 * The sets of leaves and of inner nodes 
 * of each subtopology are found.
 * Subtopologies are compared by pairs,
 * one from each topology. 
 * If they happen to have the same set of leaves
 * and of inner nodes, recombination is authorized:
 * the  subtrees that correspond with 
 * the subtopologies are swapped.
 * 
 * If more than one pair of subtopologies authorize
 * recombination, one is chosen at random.
 * A hashtable is used to memorize those pairs that do it.
 * Original topologies serves as subtopologies 
 * that authorize recombination but they do not count.
 *
 * 
 */



import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;

import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import java.awt.*;

import javax.swing.JComponent;


public class EvolFull10 extends JPanel 
                      
{
 
 private static final long serialVersionUID = 1L;

 
 
 //===========Most important parameters =========== 
 
 //The program accepts a distance matrix as input
 //else
 //it generates a random tree, calculates the distance
 //matrix associated to leaves and posits it 
 //as target matrix.
 //inoutMatrix = true: works with input matrix
 private static boolean inputMatrix = false;
 
 
 //Max number of generations to run evolution 
 static private int numberOfGenerations = 5000;
 //Report Best tree every lapse generations
 static private int lapse = 1000;


 //Number of leaves to be inserted to seed:
 //to get a tree with 5 leaves, add two leaves to seed
 //The program does not function with only one extra leaf 
 private static  int nAdditionalLeaves = 2; 


 
 //Number of topologies = number of trees that
 //conform the evolving population
 private static  int N = 50;
 

 static private long generation = 0;

 //=============================================
 

 //MUTATION PARAMETERS===========


 //Mutation of topology= 
 //   deleting leaves + reinsertion at random.
 //Choose one of two forms of deleting:
 //reverse order to that of synthesis else random.
 //One is given, the other is the complement to 1
 private static  double pReverseSynth = 0;

 //You can mutate only lengths else
 //topology + lengths
 private static double pMutateOnlyLengths = 0.5;

 //if you mutate lengths, you must decide whether or not
 //you choose the old mutation method or 
 //the new Lamarckian one with tiny mutations.
 private static double pTinyMutations = 0.5;

 //If lengths of links are mutated
 //Mutation rate per link per generation
 static private double    m = 0.5;
 //End of mutation parameters


 //Counter of right parentheses
 static private int nRightParentheses = 0;



//=================================================





 //Input distance matrix: 
 //Example 5.4, pag 99, Isaev (2001) 
 //Introduction to Mathematical Methods in Bioinformatics
 //Springer
 private static double targetMatrix[][] 
 = {{0,8,3,14,10,12},
   {8,0,9,10,6,8},
   {3,9,0,15,11,13},
   {14,10,15,0,10,8},
   {10,6,11,10,0,8},
   {12,8,13,8,8,0}};

 /*
 private static double targetMatrix5[][] 
		 = {{0,8,3,14,10},
		   {8,0,9,10,6},
		   {3,9,0,15,11},
		   {14,10,15,0,10},
		   {10,6,11,10,0}};

 private static double targetMatrix4[][] 
		 = {{0,8,3,14},
		   {8,0,9,10},
		   {3,9,0,15},
		   {14,10,15,0}};
		   
		   */

 //Number of leaves that must go in all trees > 3	
 private static	int numberOfLeaves;

 //Everything is built by recursion over this tree
 private static   String   seed = 
		    "((Aa,Ba),Ca)";
 
 //The best distance found in initialization	
 private static  int  distance0;


 private static  String leaves =  
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 private static String internalNodes = 
		 "abcdefghijklmnopqrstuvwxyz";
 //Next leaf to be grafted	
 private static  int nextLeafIndex = 2;
 //Next internal node to be inserted
 private static  int nextIntNodeIndex = 0;
 
 
 private static int maxNumberOfLeaves = 
		                      leaves.length();
 
 
 


 //Number of leaves to be deleted and reinserted
 //to create a mutation
 private static int nUndoRedo = 0;
 //Basic set of non removable leaves
 private static  String basicLeaves =   "ABC";
 //Leaves   already present in the topology
 private static  String presentLeaves =   basicLeaves;
 //Full set of leaves that conform the topology
 private static  String leavesOfTopol;
 
 private static String presentNodes =  "a";
 //Founder leaves ABC or equivalent cannot be removed
 private static String deletableLeaves = "";
 //Deleted leaves are then reinserted
 private static String reinsertingLeaves = "";
 private static String reinsertingNodes = "";

 


 //This is the original topology:
 private static  String original = "";


 

 
 //Data for binary tree
 private static String[] Record =  new String[20*N];
 

 private static  int indexOfTopologies;
 private static String topologies[] = new String[N];
 private static String trees[] = new String[N];
 private static String newTrees[] = new String[N];
 

 private static String[] bestTreeOfTopology
                             = new String[N];
 
 //Last filled position  in Record[] 
 private static int lastUsedPosition;	

 

 //Length of the link that glues a leaf to the tree
 private static int lengthsLeaves[]  = 
		                new int[maxNumberOfLeaves];
 //Length of the link that glues 
 //an inner node to another one
 private static int lengthsInnerLinks[]  = 
		               new int[maxNumberOfLeaves];
   
  
 private static Random r = new Random();
 //The control parameter of the random generator for int
 private static int maxLength = 10;
  
 //Paths from each leaf to node a
 private static String[] Paths =  
		               new String[maxNumberOfLeaves];
  
 private static String targetTree;
 private static String targetTopology;
 private static String bestTopology;
  
  

 static private double distance[] = new double[N];
 static private double fitness[] = new double[N];


 //Index of the best
 static private int best;
 //The best tree for a given topology
 static private String bestTree = "";
 //Distance from bestTree to target
 static private double bestDistance ;
 //Boolean flag for termination of job
 static private boolean workDone = true;

 

 //=========Debugging help ===============

 private static boolean print  = false;
 private static boolean print1 = false;
 private static boolean print2 = false;
 private static boolean print3 = false;
 private static boolean print4 = false;
 //Turn on print5 firstly
 private static boolean print5 = false;
 private static boolean print6 = false;
 private static boolean print7 = false;
 private static boolean print8 = true;


 //=============Strings to be drawn in pane====

 private static String titleFrame = "";
 private static String title1 = "";
 private static String title2 = "";
 private static String title3 = "";
 
 
 private static EvolFull10 newContentPane = new EvolFull10();

 private static DrawBiTree18 draw;


  //=======Constructor of main class

  public EvolFull10() 
  {
 
  }//End of constructor





public static void main(String[] args) 
{
   // Schedule a job for the event-dispatching thread:
   // creating and showing this application's GUI.
   javax.swing.SwingUtilities.invokeLater(new Runnable() 
    {
     public void run() 
     {
      //If input is a distance matrix:  
      if (inputMatrix)
      {
       printMatrix(targetMatrix, "Input target Matrix");
       boolean isD = isMatrixDistanceMatrix(targetMatrix);
         if (isD) System.out.println("\nMatrix " +
       		"is a distance matrix"); 
         else 
          {
    	   System.out.println("Matrix " +
    	       		"is NOT a distance matrix"); 
    	   System.exit(0);
           }
         numberOfLeaves = targetMatrix.length;
         nAdditionalLeaves = numberOfLeaves-3;
         presentLeaves = leaves.substring(0, numberOfLeaves);
         targetTopology = "No target topology"; 
         targetTree = "No target tree";
         leavesOfTopol =
         		leaves.substring(0, 3 + nAdditionalLeaves);
    }
    else//if a random tree is generated -> Matrix
    {      	  
     numberOfLeaves =    3 + nAdditionalLeaves;
     String topology = generateTopology();
     leavesOfTopol =
    		leaves.substring(0, 3 + nAdditionalLeaves);
     if (print6)
     System.out.println( "\nAll leaves = " + leavesOfTopol); 
     
     targetTopology = simplify(topology);
     generateLengths();
     targetTree = insertLengths(topology);
     System.out.println("Target tree: each link " +
    			"has its own length = \n" + targetTree);
     targetMatrix  =
   		     new double[numberOfLeaves][numberOfLeaves];
     targetMatrix = associatedMatrix(targetTree);
     printMatrix(targetMatrix, "Target Matrix");
    	
     double d = absDistance(numberOfLeaves,
    			              targetMatrix, targetMatrix); 
 	 
    	
     System.out.println( "\nDistance from target matrix " +
     		"to target = " + d); 
     }
 

    System.out.println("\nDefinition: \nDistance to target = "
        	+	 " Absolute distance between" 
        	+	" target matrix  and \n" 
        	+	"the distance matrix associated to " 
        	+	"the best fitting tree.\n" 
        	+	"Perfect fitting: distance = 0 \n");
     System.out.println("\nNumber of Generations = " 
        		   + numberOfGenerations);
     if (print7)
     
     System.out.println("Number of leaves = " 
                         + numberOfLeaves +
                          "\nMutation rate = "   + m +
    		             "\nPopulation size = " + N);
     
     initialization();
     evolution();
     
     }
    });
  }




//Test for being a distance matrix. 
//MaxLenght = max value among entries is calculated.
private static boolean 
    isMatrixDistanceMatrix(double[][] M)
{
 int l = M.length;
 boolean isD = true;
 //Is M(u,u) = 0?
 for(int u = 0; u < l; u++)

	if (M[u][u] ==  0) ;
	 else 
	  {
		isD = false;
		return isD;
	  }

 //Is M(u,v) = M(v,u)?
 for(int u = 0; u < l; u++)
	 for(int v = u+1; v < l; v++)
		if (M[u][v] == M[v][u] ) ;
		 else 
			{
			 isD = false;
			 return isD;
			}

 //Is M(u,v)>0 for u not v?
 for(int u = 0; u < l; u++)
	 for(int v = u+1; v < l; v++)
		if (M[u][v] >  0) ;
		 else 
			{
			 isD = false;
			 return isD;
			}

	

 //Is M(u,v) < M(u,w) + M(w,v)?
 for(int u = 0; u < l; u++)
 for(int v = 0; v < l; v++)
 for(int w = 0; w < l; w++)
	   
		if (M[u][v] <= M[u][w]+M[w][v]  ) ;
		 else 
			{
			 isD = false;
			 return isD;
			}	
 double m = 0;
 for(int u = 0; u < l; u++)
	 for(int v = u+1; v < l; v++)
		 if (M[u][v] > m ) m = M[u][v];
 maxLength = (int) m + 1;
 return true; 
}
    


//Simplification from Link notation to Newick's notation:
//(((AJ,DJ)I,BI),CI) -> (((A,D),B),C)	 	 
private static String linkToNewick(String newTree)
{
 String tree = "";
 //length of string newTree
 int l = newTree.length();
 for(int i = 0; i < l; i++ )
	  {
	   //newTree is scanned char by char	 
	  char c = newTree.charAt(i);
	  //Is c an internal node to be deleted?
	  if (charIsInternalNode(c));
	   else //add char c to output
	    tree = tree + c;
	  }
	 return tree;	 
}


//Reinitialize and generate a random topology:
//Leaves are added recursively to
//seed = ((Aa,Ba),Ca)
private static String generateTopology()
 {
  //Next leaf to be grafted	
  nextLeafIndex = 2;
  //Next internal node to be inserted
  nextIntNodeIndex = 0;
  //A tree topology is generated	
  String topology = seed;
  presentNodes = "a";
  for(int i = 0; i < nAdditionalLeaves; i++ )	 
	 topology = generateChild(topology);
  return topology;
 }
 

//Random lengths are added to a tree in Link notation 
private static void makeLengths()
 {
  //Lengths with digit values are generated
  for(int i = 0; i < numberOfLeaves; i++)
   {
	 //Zero lengths are not allowed 
	 lengthsLeaves[i] = 1 + r.nextInt(maxLength - 1); 
	 lengthsInnerLinks[i] = 1 +  r.nextInt(maxLength - 1);
   }
 } 

//The population is initialized with nTop topologies.
//Lengths are inserted to get nTop trees
//Fitness is assessed:
//Has the solution been guessed?
private static void initialization()
{
 System.out.println("\n====" +"INITIALIZATION : ");
 if (print5)
  System.out.println("\nSeed = " + seed
				 + " -> "    + linkToNewick(seed));

 if (print7)
  System.out.println( N  
		   + " random topologies " +
  		"and their trees are generated \n");

 //nTop topologies and trees are generated 
 for(int i = 0; i < N; i++)
 {
  presentLeaves = basicLeaves;
  String topo1 = generateTopology();
  topologies[i] = topo1;
  String tree = insertLengths(topo1);
  makeLengths();
  //Lengths of links are added to tree topology
  tree = insertLengths(topo1);
  trees[i] = tree;
  if (print7)
  {
	System.out.println("\n" +  i + "   " + topo1 + " -> " 
	          + linkToNewick(topo1));
	System.out.println("Tree1 = " + tree);
  } 
 }

 assessFitnesses();
 scaleFitnesses();
}


//=====================================
//============= EVOLUTION =============
//=====================================





//Does string s encode for a tree?
private static boolean isTree(String s)
{
 if (print7)
  System.out.println(
		  "Method isTree(String). String s = \n"+ s);
 boolean isTree = true;
 String numbers = "0123456789";
 presentLeaves = leaves.substring(0, numberOfLeaves);
 presentNodes = 
       internalNodes.substring(0, numberOfLeaves-2);

 String actualLeaves = "";
 String actualNodes = "";
 String leftParenthesis = "(";
 String rightParenthesis = ")";
 int counterParLeft = 0;
 int counterParRight = 0; 
 //How many times does a appear in presentNodes?

 int l = s.length();
 String c = "";
 //String s is scanned char by char
 //to get data
 for(int i = 0; i < l; i++)
  {
   c = "" + s.charAt(i);
   if (leaves.contains(c)) 
	   actualLeaves = actualLeaves + c;
   if (internalNodes.contains(c)) 
	   actualNodes = actualNodes + c;
   if (leftParenthesis.contains(c) ) counterParLeft++;
   if (rightParenthesis.contains(c) ) counterParRight++;
  }


 if (print6)
  {
   System.out.println("BEFORE REORDERING");
   System.out.println("presentLeaves length = "
			 + presentLeaves.length());
			  System.out.println("actualLeaves  length = "
			 + actualLeaves.length());
			  System.out.println("presentNodes  length  = "
			 + presentNodes.length());
			  System.out.println("actualNodes  length  = "
			 + actualNodes.length());	 
  System.out.println("presentLeaves         = "  
                                    + presentLeaves);
  System.out.println("actualLeaves with Rep = "
                                      + actualLeaves);
  System.out.println("presentNodes         = " +
		                               ""+ presentNodes);
  System.out.println("actualNodes with Rep = "
                                       + actualNodes);
  System.out.println("counterParLeft    = "+ counterParLeft);
  System.out.println("counterParRight   = "
                                   + counterParRight);
  System.out.println("nRightParentheses = "
                              + nRightParentheses);
 }



 actualLeaves = reorder(actualLeaves, leaves);
 actualNodes= reorder(actualNodes, internalNodes);

 if (print7)
  System.out.println("AFTER REORDERING");
 if (presentLeaves.length() == actualLeaves.length());
  else isTree = false;
 if (print7)
	System.out.println("isTree npresentLeaves = "+ isTree);
 if (presentNodes.length() == actualNodes.length());
  else isTree = false;
 if (print7)
  System.out.println("isTree npresentNodes = "+ isTree);


 if (print7)
  {
   System.out.println("presentLeaves length = "
	 + presentLeaves.length());
   System.out.println("actualLeaves  length = "
	 + actualLeaves.length());
   System.out.println("presentNodes  length  = "
	 + presentNodes.length());
   System.out.println("actualNodes  length  = "
	 + actualNodes.length());	 
   System.out.println("presentLeaves = "+ presentLeaves);
   System.out.println("actualLeaves  = "+ actualLeaves);
   System.out.println("presentNodes = "+ presentNodes);
   System.out.println("actualNodes  = "+ actualNodes);
   System.out.println("counterParLeft    = "+ counterParLeft);
   System.out.println("counterParRight   = "+ counterParRight);
   System.out.println("nRightParentheses = "
                              + nRightParentheses);
  }


 //Check for leaves
 if  (actualLeaves.contentEquals(presentLeaves));
  else isTree = false;

 if (print7)
		System.out.println("isTree leavesS = "+ isTree);

 //Check for nodes
 if (actualNodes.contains(presentNodes));
  else isTree = false;
 if (print7)
		System.out.println("isTree nodesS = "+ isTree);


 //Check for  parentheses
 if (counterParRight == nRightParentheses);
  else isTree = false;
 if (print7)
		System.out.println("isTree nRightPar = "+ isTree);

 if (counterParLeft == nRightParentheses);
  else isTree = false;

 if (print7)
	System.out.println("isTree nleftPar = "+ isTree);

 //Check for patterns

 //String s is scanned char by char

 String colon = ":";
 String dot = ".";
 String beforeLeaf = "(,";

 //A leaf cannot be the initial char
 if  (presentLeaves.contains("" + s.charAt(0))) 
	isTree = false;

 //Test for subpattern A:6.0a 
 for(int i = 1; i < l-6; i++)
  {
   if  (presentLeaves.contains("" + s.charAt(i)))
    {
     if ( beforeLeaf.contains("" + s.charAt(i-1)));
	     else isTree = false;   
     if ( colon.contentEquals("" + s.charAt(i+1)));
      else isTree = false;

     if (numbers.contains("" + s.charAt(i+2)));
      else isTree = false;

     if ((dot).contentEquals("" + s.charAt(i+3)));
      else isTree = false;
     if (("0").contentEquals("" + s.charAt(i+4)));
      else isTree = false;
     if  (presentNodes.contains("" + s.charAt(i+5)));
      else isTree = false;

     String node = "" + s.charAt(i+5);
     if (print7)
     System.out.println("leaf = "+ s.charAt(i)
			+ "  " + s.substring(i, i+6)  
			+ " node = " + node 
			+  "\n isTree (pattern  A:6.0a) = "+ isTree);
  }
}

	
 String beforeNode = "0)";

 //A node cannot be the initial char
 if  (presentNodes.contains("" + s.charAt(0))) 
  isTree = false;

 //Test for subpattern a:7.0
 for(int i = 1; i < l-5; i++)
  {
   if  (presentNodes.contains("" + s.charAt(i)))
    {
     if ( beforeNode.contains("" + s.charAt(i-1)));
      else isTree = false;  
    if ( colon.contentEquals("" + s.charAt(i+1)));
     else isTree = false;
    if (numbers.contains("" + s.charAt(i+2)));
     else isTree = false;
    if ((dot).contentEquals("" + s.charAt(i+3)));  
     else isTree = false;
    if (("0").contentEquals("" + s.charAt(i+4)));
     else isTree = false;
    if ((",)").contains("" + s.charAt(i+5)));
     else isTree = false;
   }
  }



 if (print7)
  System.out.println("isTree  (pattern a:7.0) = "+ isTree);

 //if pattern  )a(  appears, a must be equal to ,
 String target  = "),(";
 for(int i = 0; i < l-3; i++)
  {
   String thisChar = "" + s.charAt(i);	
   String thatChar = "" + s.charAt(i+2);
   String u = s.substring(i, i+3);
   if (  (thisChar.contentEquals(")") )
		   &  (thatChar.contentEquals("(") ) ) 		   
   { 
	 if ( target.contentEquals(u));
	  else isTree = false;
	 if (print7)
			System.out.println("u = "+ u);
   }  
  }

 if (print7)
		System.out.println("isTree )a( = "+ isTree);



 //Node a must appear tree times but others only twice
 for(int j = 0; j < presentNodes.length(); j++)
  {
   String node = "" + presentNodes.charAt(j);
   int counter = 0;
   for(int i = 0; i < l; i++)
    if ( node.contentEquals("" + s.charAt(i))) counter++;
   int mustHave;
   if   (node.contentEquals("a") ) mustHave = 3; 
    else mustHave = 2;
   if (counter == mustHave );
    else isTree = false;

   if (print7)
		System.out.println(" node = " + node 
				 +  "\n counter = " + counter	
             +  "\n mustHave = "+ mustHave);
  }

 if (print7)
		System.out.println("isTree final = "
                      + isTree);

 return isTree;
}


//Fitness is calculated for each tree
//Fitness  =
//Distance between target matrix and 
//the associated distance matrix to tree is calculated.
//The minimal distance is found: 
//this renders the fittest tree.
//Has the solution been guessed?
private static double assessFitnesses()
{
 if (print5)
  System.out.println(  "\n=====Fitnesses are assessed");
 
 String tree = "";
 //String topo = "";
 double d = 0;
 for(int i = 0; i < N; i++)
  {
   tree = trees[i];
   decoderLengths(tree);
   double M[][] = distanceMatrix(tree);
   //Absolute distance from the associated distance matrix
   //of the tree
   //to target matrix is evaluated
   d = absDistance(numberOfLeaves,targetMatrix, M); 
   distance[i] = d;

   if (print5)
    {
     System.out.println( "\nTree = \n" + trees[i]);
     System.out.println( "\nDistance to target = " + d);
    }


   if (print5)
    {
     printMatrix(M, "Distance matrix associated to tree");
     printMatrix(targetMatrix, "Target matrix");
    }

/*

//Target topology has been found?
topo = deleteLengths(bestTree);
topo = simplify(topo);
if (targetTopology.contentEquals(topo))
 {
  System.out.println( "\n \nTOPOLOGY HAS BEEN FOUND: \n" +
			"at generation " + generation
			+ "\nFound topology  = \n" + topo
			+ "\nTarget Topology = \n" + targetTopology);
 //System.exit(0);
 if (print5) print5 = false;
}

*/  
  
 //End work if solution has been found.
 if (d == 0) 
  {
   System.out.println("\n BestTree  = \n" +  bestTree    
		           + "\nDistance = " + bestDistance 
		           + "\ntarget tree = \n" + targetTree 
		           + "\nInitial distance = " + distance0);

   workDone = false;
  }
 }//End of for(int i...




 //The fittest individual minimizes distance function
 bestDistance = distance[0];
 best = 0;
 for(int i = 1; i < N; i++)
 {
  if (distance[i] < bestDistance )
   {
    bestDistance = distance[i];
    best = i;
   }
 }

 //Best tree 
 bestTree = trees[best];
 
 
 
 
 
 if (generation == 0 ) distance0 = (int) bestDistance;
 if (print8)
  {
   System.out.println("\n BestTree  = \n" +  bestTree    
		           + "\nDistance = " + bestDistance 
		           + "\ntarget tree = \n" + targetTree 
		           + "\nInitial distance = " + distance0);
   //Target topology has been found?
   String topo = deleteLengths(bestTree);
   topo = simplify(topo);

   System.out.println( "\nBest topology  = \n" + topo
			+ "\nTarget Topology = \n" + targetTopology);
   if (inputMatrix);
   else
   {   
   if (targetTopology.contentEquals(topo))
    {
     System.out.println( "\n\nTOPOLOGY HAS BEEN FOUND!\n");
    
    }
   }
  }
 return bestDistance;
}



//Fitness is rescaled to take values
//in within 0(the worst) and 1(the best)
private static void scaleFitnesses()
{
 if (print5)
  System.out.println(  "\n====Fitnesses are rescaled: " +
 		"0(the worst) and 1(the best)");
 double maxD = distance[0];
 double minD = distance[0];
 double b = 1 ;
 //Min and max distances are found
 for(int i = 1; i < N; i++)
  {
	if ( distance[i] > maxD ) maxD =  distance[i];
	if ( distance[i] < minD ) minD =  distance[i];
  }
	 
 if (minD == maxD) //All   equally fit
  {
   for(int i = 0; i < N; i++)
	 fitness[i] = 1;
  }
 else
  {
	 //Distance to target determine fitness:
	//fitness[maxD] = 0;  	 fitness[minD] = 1;
	 for(int i = 0; i < N; i++)
	 {  
	  b = 1;
	  b = b/(minD - maxD);
	  double a = -b*maxD;
	  fitness[i] = a + b * distance[i];	 
	}
  }
 if (print5)
  for(int i = 0; i < N; i++)
   {
	 System.out.println("Individual = "
             + i + " has distance " +
          distance[i] + " and fitness " + fitness[i]);
   }
}

//Probability of cloning is proportional to fitness.
//N new individuals are formed. 
private static void selectiveCloning()
{
 if (print5)
  System.out.println(  "\n====Selective cloning");
 //The best is cloned
 newTrees[0] = trees[best];
 if (print5)
	 System.out.println("Individual = "
                      + best + " with fitness " +
			 fitness[best] + "-> new individual " + 0
			 + " = "  + newTrees[0]  );	
 //N-1 children are sampled
 for(int i = 1; i < N; i++)
  {
   //flag
   boolean go = true;
   while (go)
    {
	 //An individual is chosen at random  
	 int k = r.nextInt(N);
	 //Probability of cloning is proportional to fitness. 
	 double q = r.nextDouble();	 
	 if ( q < fitness[k]) //make a copy of k
	  {	    
		 newTrees[i] = trees[k];		  
		 if (print5)
			 System.out.println("Individual = "
		                      + k + " with fitness " +
					 fitness[k] + "-> new individual " + i
					 + " = "  + newTrees[i]  );		 
		 go = false;//end task		  
	  }
     }

  }//End of for //N-1 children are sampled
 }

//============Mutation============


//String s ( a char) is deleted from mother
public static String delete(String s, String mother)
{
 String output = "";
 int l = mother.indexOf(s);
 output = mother.substring(0, l)  
		 + mother.substring(l+1) ;
 return output;
}



//String leaves is reversed
public static String reverse(String leaves)
{
 String reversed = "";
 int l = leaves.length();
 for(int i = l-1; i >= 0; i--)
  {
	String c = "" + leaves.charAt(i); 
	reversed = reversed + c;
  }
 if (print1)
 System.out.println("Reverse order = " + reversed);
 seed =  "((Za,Ya),Xa)";
 if (print1)
 System.out.println("seed = " + seed);
 basicLeaves = "ZYX";
 presentLeaves =   basicLeaves;
 return reversed;
}


//String leaves is permuted
public static String permuteOrder(String leaves)
{
 if (print1)
  System.out.println("original and permuted order = \n" 
                        + leaves);
 String permuted = "";
 int l = leaves.length();
 int m = l;
 for(int i = 0; i < l; i++)
  {
	int d = r.nextInt(m) ;
	String c = "" + leaves.charAt(d); 
	leaves = delete(c,leaves);
	m = m-1;
	permuted = permuted + c;
  }
 if (print1)
 System.out.println(permuted);
 String first = "" + permuted.charAt(0);
 String second = "" + permuted.charAt(1);
 String third = "" + permuted.charAt(2); 
 seed = "((" +first + "a," +second + "a)," +third + "a)";
 if (print1)
  System.out.println("seed = " + seed);

  basicLeaves =  first + second + third;
 presentLeaves =   basicLeaves;
 
 return permuted;	
}


//Detect internal nodes, different than leaves
private static boolean charIsInternalNode(char c)
{
 String s = "" + c;	 
 if (internalNodes.contains(s)) return true;
  else return false;
}

//Simplification from Link notation to Newick's notation:
// (((AJ,DJ)I,BI),CI) -> (((A,D),B),C)
private static String simplify(String newTree)
{
 String tree = "";
 //length of string newTree
 int l = newTree.length();
 for(int i = 0; i < l; i++ )
  {
   //newTree is scanned char by char	 
   char c = newTree.charAt(i);
   //Is c an internal node to be deleted?
   if (charIsInternalNode(c));
    else //add char c to output
   tree = tree + c;
  }
 return tree;	 
}


//Rule 2 is used to generate a new tree:
//(S)I -> ((S)N,FN)I.
//S = subtree, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string seedTree
private static String useRule2(String seedTree, int i, 
    String F, String N)
{
 String newTree = "";
 char I = seedTree.charAt(i);
 if (print1)
  System.out.println("i = " + i + " I = " + I);
 //Subtree S of link (S)I must be read 
 String S = "";
 //Counter of left seedhesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = i;
 boolean go = true;
 //All right seedheses to the near left of I are captured
 while (go)
 {
  position = position -1;
  String a = "" + seedTree.charAt(position);

  if (print1)
   System.out.println("a = " + a + " p = " + pr);
  if (a.equals(pr) ) 
   {
    counter++;
	 S = a + S;
   }
   else go = false;
 }//End of while
 if (print1)
 System.out.println("counter = " + counter);
 //The complete subtree is captured:
 //seedheses must balanced
 while (counter > 0)
 {   
  String a = "" + seedTree.charAt(position);
  if (a.equals(pr))   counter++;
  if (a.equals(pl))  counter--;
  S = a + S;
  if (print1)
   System.out.println("Subtree = " + S);
  position = position -1;
 }
 //Replacement rule 2:
 //((S)n,Fn)i -> (S)i
 //where   F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =   S  + I; 
 String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
 //Rule 1 is executed
 newTree = seedTree.replace(SnFni, replacement);		
 //If I links to a complex structure
 return newTree;	 
}



//Rule 2 is used to generate a new tree:
//(S)I -> ((S)N,FN)I.
//S = subtree, I = internal node, F = new leaf, 
//N = new internal node.
//Rule 2 grafts new branch FN to link (S)I
//i is the position of char I in string seedTree
private static String useRule2(String seedTree, int i, 
      char F, char N)
{
 String newTree = "";
 char I = seedTree.charAt(i);
 if (print1)
  System.out.println("i = " + i + " I = " + I);
 //Subtree S of link (S)I must be read 
 String S = "";
 //Counter of left seedhesis
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = i;
 boolean go = true;
 //All right seedheses to the near left of I are captured
 while (go)
  {
   position = position -1;
   String a = "" + seedTree.charAt(position);
 
   if (print1)
    System.out.println("a = " + a + " p = " + pr);
   if (a.equals(pr) ) 
    {
     counter++;
	 S = a + S;
    }
   else go = false;
  }//End of while
 if (print1)
 System.out.println("counter = " + counter);
 //The complete subtree is captured:
 //seedheses must balanced
 while (counter > 0)
  {   
   String a = "" + seedTree.charAt(position);
   if (a.equals(pr))   counter++;
   if (a.equals(pl))  counter--;
   S = a + S;
   if (print1)
   System.out.println("Subtree = " + S);
   position = position -1;
  }
 //Replacement rule 2:
 //((S)n,Fn)i -> (S)i
 //where   F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =   S  + I; 
 String	replacement =  "("+ S +   N + "," 
		          + F + N+ ")" + I;
 //Rule 1 is executed
 newTree = seedTree.replace(SnFni, replacement);		
 //If I links to a complex structure
 return newTree;	 
}

//Input: seedTree, a tree in Newick's notation.
//Output : newTree =  seed tree + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String generateChild(String seedTree) 
{
	
 nextLeafIndex++;
 char nextLeaf = leaves.charAt(nextLeafIndex);

 presentLeaves = presentLeaves + nextLeaf;
 if (print5)
  System.out.println("Method generateChild \n " +
		"Present leaves = " + presentLeaves);
 //Next internal node to be inserted
 nextIntNodeIndex++;
 char nextIntNode = internalNodes.charAt(nextIntNodeIndex);
 presentNodes = presentNodes + nextIntNode;
	
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new tree generated by replacement
 String newTree = "";
 //length of seed tree
 int l = seedTree.length();
 //counter of links = 0;
 int counter = 0;
 //Links are counted
 for(int i = 0; i < l; i++ )
  {
   //seedTree is scanned char by char	 
   char I = seedTree.charAt(i);
   //Is I an internal node that denotes a link?
   if (charIsInternalNode(I)) counter++;	
  }
 //A link is chosen 
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
 {
  //seedTree is scanned char by char	 
	char I = seedTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = seedTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(seedTree, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTree = seedTree.replace(link, replacement);		
		 }
		 if (print1)
		System.out.println(newTree + " -> " 
		                       + simplify(newTree));			

	  
		 i = l;
	   }
	  counter++;	
	} 
 } 
 return newTree;
}


//Input: seedTree, a tree in Newick's notation.
//Output : newTree =  seed tree + a new link.
//The new link is added by a generative grammar.
//A random election is made among all possibilities.
private static String reinsertLeaf(String seedTree) 
{
 nextLeafIndex++;
 char nextLeaf = reinsertingLeaves.charAt(nextLeafIndex);
 presentLeaves = presentLeaves + nextLeaf;
 if (print6)
 System.out.println("Method reinsertLeaf \n " +
		"Present leaves = " + presentLeaves);
 //Next internal node to be inserted
 nextIntNodeIndex++;
 char nextIntNode = 
		 reinsertingNodes.charAt(nextIntNodeIndex);
 presentNodes = presentNodes + nextIntNode;
	
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new tree generated by replacement
 String newTree = "";
 //length of seed tree
 int l = seedTree.length();
 //counter of links = 0;
 int counter = 0;
 //Links are counted
 for(int i = 0; i < l; i++ )
 {
  //seedTree is scanned char by char	 
  char I = seedTree.charAt(i);
  //Is I an internal node that denotes a link?
  if (charIsInternalNode(I)) counter++;	
 }
 //A link is chosen 
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
 {
  //seedTree is scanned char by char	 
	char I = seedTree.charAt(i);
	//Is I an internal node that denotes a replacement?
	if (charIsInternalNode(I))
	 {
	  if (counter == k) //Is I the chosen link?
	   {
		//Check for the char before I	
		char a = seedTree.charAt(i-1);
		//If I links to a complex structure
		if (a == ')') newTree = useRule2(seedTree, i, 
				               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		 replacement = "(" + a + nextIntNode + "," 
			          + nextLeaf + nextIntNode + ")" + I;
		 //Rule 1 is executed
		 newTree = seedTree.replace(link, replacement);		
		 }
		 if (print1)
		System.out.println(newTree + " -> " 
		                       + simplify(newTree));			

	  
		 i = l;
	   }
	  counter++;	
	} 
 } 
return newTree;
}


//Leaf F and node n conform  link Fn, which is  deleted.
//We apply the antirule1:
//(An,Fn)i  -> Ai
private static String  antiRule1(String topology, 
		int positionLeaf, int positionNode)
{
 if (print7)
  System.out.println("AntiRule1:  input topology = \n" 
                      + topology);
 String newTopology = "";
 int beginIndex = positionLeaf - 4;
 int endIndex = positionNode + 3;
 
 int l = topology.length();
 if (endIndex < l)
 {
  String AnFni = topology.substring(beginIndex, endIndex);
  if (print7)
  System.out.println("Anfni = " + AnFni);
  String Ai = "" + topology.charAt(beginIndex+1)
              + topology.charAt(endIndex-1);
  String detector = "" + Ai.charAt(1);
  if (detector.contentEquals(","))
   Ai = "" + Ai.charAt(0) + topology.charAt(positionNode)
       + detector;
  
  if (print7)
  System.out.println("Ai = " + Ai);
  newTopology = topology.replace(AnFni, Ai);
  
  if (print7)
   System.out.println("Output topology = \n" + newTopology);
  }
 else //Pattern //((S),Fn) -> (S)
  {
	//Leaf to be deleted	
	 String leaf = "" + topology.charAt(positionLeaf);
	//Accompanying inner node  
	 String innerNode = "" 
		            + topology.charAt(positionNode);
	 newTopology = antiRule2a(topology,  leaf, innerNode);
  }
 return newTopology;
}


//Leaf F is deleted from tree when accompanying node is a
//and F is in a terminal position.
//We apply the antirule2a:
//((S),Fn) -> (S)
private static String  antiRule2a(String topology, 
		String leaf, String node)
{
 int l = topology.length();
	String newTopology = topology.substring(1, l-4);
	if (print7)
		 System.out.println("AntiRule2a newTree = \n"
				 + newTopology);	
	//System.exit(0);
 return newTopology;
}


//Leaf F appeared as an insertion 
//into a complex structure S.
//It is deleted.
//We apply the antirule2:
//((S)n,Fn)i -> (S)i
//positionNode refers to the last appearing of node n
private static String  antiRule2(String topology, 
		int positionLeaf, int positionNode)
{
 String newTopology = "";	
 if (print1)
		 System.out.println("AntiRule2");
 //Leaf to be deleted	
 String leaf = "" + topology.charAt(positionLeaf);
 if (print1)
  System.out.println("\nDeletion = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
	            + topology.charAt(positionNode);
 int m = topology.length();
 int p = m-1;
 if (p == positionNode) //Case ((S),Fn), F = leaf to delete
	 newTopology = antiRule2a(topology,  leaf, innerNode);
 else
 {
	 
 int firstPosition = topology.indexOf(innerNode);

 String i = ""; 
 if (positionNode +2 < topology.length())
       i = ""+ topology.charAt(positionNode+2);
 if (print1)
  System.out.println("FirstPosition = " + 
                    firstPosition + " i = " + i);
 //Subtopology S of link (S)I must be read 
 String S = "";
 //Counter of left parentheses
 int counter = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Position to be scanned
 int position = firstPosition;
 boolean go = true;
 //All right parentheses to the near left of I are captured
 while (go)
 {
  position = position -1;
  String as = "" + topology.charAt(position);
 
  if (print1)
  System.out.println("a = " + as + " p = " + pr);
  if (as.equals(pr) ) 
  {
	counter++;
	S = as + S;
  }
  else go = false;
 }//End of while
 if (print7)
  System.out.println("counter = " + counter);
 //The complete subtopology is captured:
 //parntheses must balanced
 while (counter > 0)
  {   
   String as = "" + topology.charAt(position);
   if (as.equals(pr))   counter++;
   if (as.equals(pl))  counter--;
   S = as + S;
   if (print7)
    System.out.println("Subtopology = " + S);
   position = position -1;
  }
 String n = innerNode;
 //Replacement antiRule 2:
 //((S)n,Fn)i -> (S)i
 //where A and F are leaves, n and i are  inner nodes,
 //(S) is a complex structure. 
 String SnFni	 =  "("+ S  + n + "," + leaf + n + ")" + i ; 
 String	Si = S + i;
 if (print7)
  System.out.println("\nSnFni = " + SnFni 
		  + "\n Si = " + Si);
 //Rule 1 is executed
 newTopology = topology.replace(SnFni, Si);	
 if (print7)
  System.out.println("Output topology = \n" + newTopology);
 }
 return newTopology;
}


//Leaf F and node n conform  link Fn  and are deleted.
//We apply   antirule3:
//(Fn,An)i  -> Ai
private static String  antiRule3(String topo, 
		int positionLeaf, int positionNode)
{
 
 if (print7)
  System.out.println("AntiRule3"
		+ "\nInput topology = \n" + topo);
 String node = "" + topo.charAt(positionNode);
 int beginIndex = positionLeaf - 1;
 int endIndex = positionNode + 6;
 String FnAni = topo.substring(beginIndex, endIndex);
 if (print7)
 System.out.println("FnAni = " + FnAni);
 String Ai = "" + topo.charAt(beginIndex+4)
            + topo.charAt(endIndex-1);
 if (print7)
  System.out.println("Ai = " + Ai);
	
 String lastChar = "" + Ai.charAt(1);
 //(Da,Ba)d or (Da,Ba)?
 if (lastChar.contains(","))
  Ai = "" + Ai.charAt(0) + 	node + lastChar;
 
 topo = topo.replace(FnAni, Ai);	
 
 
 if (print7)
  System.out.println("Output topology = \n" + topo);
 return topo;
}



//Leaf F appeared as an insertion into a complex structure S.
//it is deleted.
//We apply antirule4:
//(Fn,(S)n)i -> (S)i
private static String  antiRule4(String topology, 
		int positionLeaf, int positionNode)
{

 //Leaf to be deleted	
 String leaf = "" + topology.charAt(positionLeaf);

 //Accompanying inner node  
 String innerNode = "" 
		            + topology.charAt(positionNode);
 if (print7)
  System.out.println("\nAntiRule4"
		+ "\nLeaf to be deleted = " + leaf
		 + " at position " + positionLeaf
		+  "\nAccompanying node = " + innerNode
			 + " at first position " + positionNode);
 //Second position of accompanying node
 int positionNode2 = 0;
 int l = topology.length();
 boolean go = true;
 //topology is scanned char by char
 for(int i = positionNode + 1; i < l; i++ )
  {
	 String c = "" + topology.charAt(i);
	
	 if (go)
	 {
	  if (c.equals(innerNode))
		 {
		  positionNode2 = i;	
          go = false;
          if (print7)
        	System.out.println(
        		"Second position of accompanying node = " 
        		+ positionNode2 + " i = " + i);
		 }
	 }	
  }

	
 int beginIndex = positionLeaf - 1;
 int endIndex = positionNode2+2;
 String i = ""; 
 if (endIndex < topology.length())
	         i = ""+ topology.charAt(endIndex);
	 
 if (print7)
  System.out.println("EndIndex = " + 
	                      endIndex + " i = " + i);
	 
 String FnSni = 
		 topology.substring(beginIndex, endIndex) + i;
 if (print7)
  System.out.println("FnSni = " + FnSni);
 String S = topology.substring(beginIndex+4, endIndex-2);
 if (print7)
	System.out.println("S = " + S);
 String	Si = S + i;
 if (print7)
	 System.out.println("Si = " + Si);
	 
 String newTopology = topology.replace(FnSni, Si);
 if (print7)
	 System.out.println("Output topology = \n" 
                        + newTopology); 
 //System.exit(0);	 
 return newTopology;
}




//The last inserted  leaf is deleted from topology.
private static String deleteLastLeaf(String topology) 
{
 //Leaf to be deleted	
 String leaf = "" + leaves.charAt(nextLeafIndex);
 if (print1)
 System.out.println("\nLeaf to be deleted = " + leaf);
 //Accompanying inner node  
 String innerNode = "" 
      + internalNodes.charAt(nextIntNodeIndex);
 if (print1)
  System.out.println("Accompanying inner  node = " 
              + innerNode);		
 //new topology = output
 String newTopology = "";
 //length of  topology as string
 int l = topology.length();
 //Last positions are found
 int positionLeaf = -1;
 int positionNode = -1;  
 //topology is scanned char by char
 for(int i = 0; i < l; i++ )
  {
	 String c = "" + topology.charAt(i);
	 if (c.equals(leaf)) positionLeaf = i;
	 if (c.equals(innerNode)) positionNode = i;
  }
 if (print1)
 System.out.println("Position of " + leaf + 
		                   " = " + positionLeaf);
 if (print1)
  System.out.println("Position of " + innerNode + 
    " = " + positionNode);
 int firstPosition = topology.indexOf(innerNode);
 //Check for the char before firstPosition	
 char a = topology.charAt(firstPosition - 1);
 //If I links to a complex structure
 if (a == ')')
  newTopology = 
     antiRule2(topology,positionLeaf,positionNode);
 else  //Apply antiRule1 for simple cases
	  newTopology = 
	    antiRule1(topology,positionLeaf,positionNode);
 nextLeafIndex--;
 nextIntNodeIndex--;
 return newTopology;	
}


//A random leaf is deleted from topology. 
private static String deleteRandomLeaf(String topology) 
{
 if (print7)
  {
   System.out.println("\nMthd deleteRandomLeaf: topology = "
           +  "\n" + topology + "\ndeletableLeaves = " 
		   + deletableLeaves);
  }
  
  
 int l = deletableLeaves.length();
 int position = r.nextInt(l);
 //Leaf to be deleted	
 String leaf = "" + deletableLeaves.charAt(position);
 int nodeIndex = topology.indexOf(leaf) + 1;
 //Accompanying inner node  
 String innerNode = "" 
          + topology.charAt(nodeIndex);

 deletableLeaves = delete(leaf,deletableLeaves);
 presentLeaves = delete(leaf,presentLeaves);
 reinsertingLeaves = reinsertingLeaves + leaf;
 reinsertingNodes = reinsertingNodes + innerNode;


 if (print7)
 System.out.println("\nLeaf to be deleted = " + leaf
		+ "\nReinserting leaves = " + reinsertingLeaves
       + "\nAccompanying inner  node = "   + innerNode
       + "\nReinserting nodes = " + reinsertingNodes);

 //new topology = output
 String newTopology = "";
 //length of  topology as string
 l = topology.length();
 //Last positions are found
 int positionLeaf = topology.indexOf(leaf);
 int positionNode = positionLeaf + 1;  


 if (print7)
  System.out.println("Position of " + leaf 
		        +  " = " + positionLeaf
		        + "\nPosition of " + innerNode  
                + " = " + positionNode);
 //Which set of deleting rules to apply:
 // 1 and 2 else 3 and 4?
 String previousChar = "" + topology.charAt(positionLeaf-1);
 if (previousChar.equals(",")) //Apply antiRules 1 or 2
  {
   
   int firstPosition = topology.indexOf(innerNode);
   //Check for the char before firstPosition	
   char a = topology.charAt(firstPosition - 1);
   //If I links to a complex structure
   if (a == ')')
   {
	if (print7)
	 System.out.println("Apply antiRule 2");
    newTopology = antiRule2(topology,positionLeaf,
    		                positionNode);
   }
   else  //Apply antiRule1 for simple cases
	{
	  if (print7)
		System.out.println("Apply antiRule 1");
	  newTopology = antiRule1(topology,positionLeaf,
			                  positionNode);
	}
   nextLeafIndex--;
   nextIntNodeIndex--;
  }
 else //Apply antiRules 3 or 4
  {
   int firstPosition = topology.indexOf(innerNode);
   //Check for the char before firstPosition	
   char a = topology.charAt(firstPosition + 2);
   //If I links to a complex structure
   if (a == '(')
    {
	 if (print7)
	  System.out.println("Apply antiRule 4");
	 newTopology = antiRule4(topology,positionLeaf,
			                  positionNode);
     }
     else  //Apply antiRule3 for simple cases
	 {
	  if (print7)
	  System.out.println("Apply antiRule 3");
      newTopology = antiRule3(topology,positionLeaf,
    		                  positionNode);
	 }
   nextLeafIndex--;
   nextIntNodeIndex--;	 
  }
 if (print7)
  System.out.println("topology after deletion \n" 
                                   + newTopology);
 
 return newTopology;	
}



//Mutation = 
//Random deletions are executed and then  reinserted.
public static String  mutationRandomDeletion(
		           String topo, int nAdditionalLeaves)
{
	//Reinitialization
 reinsertingLeaves = "";
 reinsertingNodes = "";
 //Some leaves are deleted: random order
	 if (print7)
 System.out.println("Some leaves are deleted AT RANDOM");
 for(int i = 0; i < nUndoRedo; i++ )	 
	{
	 //Random leaf is deleted.
	 topo = deleteRandomLeaf(topo);
	 }
 //Deleted leaves are reinserted
 nextLeafIndex = -1;
 nextIntNodeIndex = -1;
 if (print7)
  System.out.println("Actual topology = " + topo
  + "\nReinserting leaves = "  + reinsertingLeaves
  + "\nReinserting nodes = " + reinsertingNodes); 
 
 
 //How many times does a appear in topo?
 int howManya = 0;
 String c = "";
 for(int j = 0; j < topo.length(); j++)
  {
   c = "" + topo.charAt(j);
   if (c.contentEquals("a"))
	  howManya++;  
  }
 if (howManya == 3) 
  {

   int t = presentNodes.length();
   reinsertingNodes =  reinsertingNodes.
		replace("a", "" + internalNodes.charAt(t)); 
   if (print7)
	   System.out.println("New Reinserting nodes = " 
                 + reinsertingNodes);
  }
 
 
 for(int i = 0; i < nUndoRedo; i++ )	 
	{
	 topo = reinsertLeaf(topo);
	}
 if (print7)
  System.out.println("Topology after del+reinsertion" +
	  		"  = \n"  + topo); 
 return topo;
}

//Two leaves, different than A, 
//are chosen at random to be swapped.
public static String swapLeaves(String topology)
{	
 presentLeaves = reorder(presentLeaves, leaves);
 //A,B and C are untouchable
 presentLeaves = presentLeaves.substring(3);
 int n = presentLeaves.length();
 int i = r.nextInt(n);
 String s1 = "" + presentLeaves.charAt(i);
 int j = i;
 String s2 = "";
 while (j==i)
  {
	j = r.nextInt(n);
	s2 = "" + presentLeaves.charAt(j);	 
  }
 
 String output = topology.replace(s1, "=");
 output = output.replace(s2, s1);
 output = output.replace("=", s2);
 if (print6)
	{
	 System.out.println("s1 = " + s1 + " s2 = " + s2);
		System.out.println("\nMethod swapLeaves " 
	                       + "\nInput topology =   "
		                   + topology
		                   + "\nOutput topology =  "
		                   + output);	;	
	}
 return output;
}


//Input = string s. 
//Output = an alphabetically ordered list 
//of the chars of s
//according to the given alphabet. 
//Repetitions are forgotten.
//Example : BACDA -> ABCD
private static String reorder(String s, String alphabet)
{
 String myLeaves = "";
 int n = alphabet.length();
 for(int j = 0; j < n; j++)
  {
   String c = "" + alphabet.charAt(j);
   if (s.contains(c))
	   myLeaves = myLeaves + c;
  }	
 if (print5)
 System.out.println("Reorder " + s + " " + myLeaves);
 return myLeaves;
}


//Insert link = (leaf,node) into oldtopology
private static String insertLink(String oldtopology, 
		String nextLeaf, String nextIntNode )
{
 // nextLeafIndex++;
 // presentLeaves = presentLeaves + nextLeaf;
 if (print7)
	 System.out.println("Method insertLink \n" +
			"leaf = " + nextLeaf
			+"\nNode = " + nextIntNode);
	 
 //link to be replaced
 String link = "";
 //String link is replaced by what:
 String replacement = "";
 //new topology generated by replacement
 String newTopology = "";
 //length of seed topology
 int l = oldtopology.length();
 //counter of links = 0;
 int counter = 0;
 //Nodes are counted
 for(int i = 0; i < l; i++ )
  {
   //oldtopology is scanned char by char	 
   char node = oldtopology.charAt(i);
   //Is I an internal node that denotes a link?
   if (charIsInternalNode(node)) counter++;	
  }
 //Place for insertion?
 int k = r.nextInt(counter);
 //System.out.println("Random number = " + k);
 //new branch is grafted at k
 counter = 0;  
 for(int i = 0; i < l; i++ )
  {
   //oldtopology is scanned char by char	 
   char I = oldtopology.charAt(i);
   //Is I an internal node that denotes a replacement?
   if (charIsInternalNode(I))
	{
	 if (counter == k) //Is I the chosen link?
	 {
	  //Check for the char before I	
	  char a = oldtopology.charAt(i-1);
	  //If I links to a complex structure
	  if (a == ')') 
				newTopology = useRule2(oldtopology, i, 
					               nextLeaf, nextIntNode);
		else //if I links to a leaf, use rule 1
		 {
		  //Link to be replaced  
		  link = "" + a + I; 
		  //Replacement rule 1:
		  //aI -> (aN,FN)I;   N= nextIntNode; 
		  //F = nextLeaf
		  replacement = "(" + a + nextIntNode + "," 
				       + nextLeaf + nextIntNode + ")" + I;
		  //Rule 1 is executed
		  newTopology = oldtopology.replace(link, 
					                         replacement);		
		  }
		if (print1)
			System.out.println(newTopology + " -> " 
			                + simplify(newTopology));			

		  
		i = l;
	   }
	 counter++;	
	} 
  } 
 return newTopology;	
}


//The complement of a  chosen at random subtopology
//is restructured
public static String  
       topComplementReestructuring(String topology)
{
 if (print7)
	System.out.println(	"\nMMethod subtopReestructuring"
			+ "\nTopology = \n" + topology); 
 //Data about topology
 //Leaves and nodes in topology
 int lTop = topology.length();
 String leavesOfTopol = "";
 String nodesOfTopol = "";
 String c = "";	
 //Number of right parenthesis
 nRightParentheses = 0;
 String rightPar = ")";
 for(int j = 0; j < lTop; j++)
  {
	 c = "" + topology.charAt(j);
	 if (leaves.contains(c)) 
	 	  leavesOfTopol = leavesOfTopol + c; 
	 if (internalNodes.contains(c))
	 	   nodesOfTopol = nodesOfTopol + c; 
	 if ( c.contentEquals(rightPar)) nRightParentheses++;	    
  }

 //A subtopology is chosen:
 //it will remain untouched while its
 //complement is restructured
 String newTopology = "";			  
 String[] subtop = new String[counterSubtop];
 subtop = subtopologies(topology);
 int i;
 if (counterSubtop > 1)
  i = r.nextInt(counterSubtop-1) + 1;
  i=1;
 String subTop = subtop[i];
 if (print7)
  System.out.println("\nChosen subtopology = \n" + subTop);
 
 int l = subTop.length();

 String leavesOfSubTop = "";
 String nodesOfSubTop = "";
 boolean hasA= false;
 boolean hasB = false;
 boolean hasC = false;
 
 //String subTop is scanned char by char
 //to get information
 for(int j = 0; j < l; j++)
  {
   c = "" + subTop.charAt(j);
   if (leaves.contains(c)) 
	   leavesOfSubTop = leavesOfSubTop + c;
   if (internalNodes.contains(c)) 
	   nodesOfSubTop = nodesOfSubTop + c; 
   if (c.contentEquals("A")) hasA = true;
   if (c.contentEquals("B")) hasB = true;
   if (c.contentEquals("C")) hasC = true;
  }
 //How many times does a appear in nodesOfSubTop?
 int howManya = 0;
 for(int j = 0; j < nodesOfSubTop.length(); j++)
 {
  c = "" + nodesOfSubTop.charAt(j);
  if (c.contentEquals("a"))
	  howManya++;  
 }

 int nleavesOfSubTop = leavesOfSubTop.length();
 //All large subtopologies are marked with 4
 if (nleavesOfSubTop > 3) nleavesOfSubTop = 4;
 
 if (print7)
	 System.out.println("\nnleavesOfSubTop = "
                    + nleavesOfSubTop);
 String newSubTop = "";
  //Restruc. of complement of large subtopologies 
  
   //Lacking leaves and nodes of subTop
   String lackingLeaves = leavesOfTopol;
   l = leavesOfTopol.length();
   for(int j = 0; j < l; j++)
	{
	 c = "" + leavesOfTopol.charAt(j);
	 if (subTop.contains(c)) 
	 	   lackingLeaves = lackingLeaves.replace(c, "");     
	}
	  
   String lackingNodes = nodesOfTopol;
   l = nodesOfTopol.length();
   for(int j = 0; j < l; j++)
    {
	 c = "" + nodesOfTopol.charAt(j);
	 if (subTop.contains(c)) 
	 	   lackingNodes = lackingNodes.replace(c, "");     
	}
	 lackingLeaves = reorder(lackingLeaves, leaves);
	 lackingNodes = reorder(lackingNodes, internalNodes);
   if (print7)  
   System.out.println("Case large subTop"
        + "\nleavesOfSubTop  = " + leavesOfSubTop
	 	+ "\nnodesOfSubTop = " + nodesOfSubTop 
	 	+ "\nleavesOfTopol = " + leavesOfTopol
	 	+ "\nlacking leaves in order = " + lackingLeaves
	 	+ "\nlacking nodes in order= " + lackingNodes);
	
   
   //Our restructuring method needs more than 
   //2 lacking leaves
   if ((lackingLeaves.length() < 2) 
		   || (lackingNodes.length() < 1)) 
   {
	 newSubTop = subTop; //Do nothing
	 newTopology = topology; 
   }
   else //Rearrange the complement of subTop 
    {
	 //Compose a topology with 
	 //lackingLeaves and subTop 
     //but touch not subtop. To that aim,
	 //replace it by the  symbol ?
	 //and later restore it back to subtop:
	   
	   
	 
	 //  Step 1: a basic topology is built.
	 //Begin with ((A,B),C) and 
	 //replace A or B or C by ? appropriately.	 
	 String basicTopology = "";
	   
	 //If subTop has A  
	 if (hasA) 
	 {
	  if (( (hasA) & (!(hasB))) & (!(hasC))) //Has only A
	  {
	   if (howManya == 0) //Has no a
				basicTopology = "((?a,Ba),Ca)";
	   if (howManya == 1)	//Has one a
				basicTopology = "((?, Ba),Ca)";
	  }
	  if ((hasB) & (!(hasC)))	//has A and B
	  {
	   basicTopology = "(?,Ca)";  
	  }
	 }
	//If subTop has only B
	 if (( (hasB) & (!(hasA))) & (!(hasC))) 
	   {
		   if (howManya == 0) //Has no a
				basicTopology = "((Aa,?a),Ca)";
		  if (howManya == 1)	//has A and B
				basicTopology = "((Aa,?),Ca)";		   
	   }   
		   
		 
	 //If subTop has only C  
	 if (( (hasC) & (!(hasA))) & (!(hasB))) 	 
	   {
		   if (howManya == 0) //Has no a
				basicTopology = "((Aa,Ba),?a)";
		  if (howManya == 1)	//has a
				basicTopology = "((Aa,Ba),?)";		   
	   }
	 
	 //If subtop has no A, no B, no C
	 if (( (!(hasC)) & (!(hasA))) & (!(hasB)))
	   {
		//Add ? as a new "leaf"
		 lackingLeaves = lackingLeaves + "?"; 	   
	   }

	 
	   lackingLeaves = lackingLeaves.replace("A", "");
	   lackingLeaves = lackingLeaves.replace("B", "");
	   lackingLeaves = lackingLeaves.replace("C", "");
	   lackingNodes = lackingNodes.replace("a","");
	   

	 //Basic topology is done. Now:
	 //Insert lacking leaves (but do not touch subTop)
	 newSubTop = basicTopology;
	 while(! (lackingLeaves.contentEquals("")))
	  {
	   if (print7)
		System.out.println("\nnewSubTop e = " + newSubTop
				+ "\nlacking leaves  = " + lackingLeaves
				+ "\nlacking nodes = " + lackingNodes);	   
			   	   
		 String leaf = "" + lackingLeaves.charAt(0);
	   String node = "" + lackingNodes.charAt(0);
	   newSubTop = insertLink(newSubTop, leaf, node );
	   lackingLeaves = lackingLeaves.substring(1);
	   lackingNodes = lackingNodes.substring(1);	
	  }
	 
	 //Restore subTop
	 newTopology = newSubTop.replace("?", subTop);
	
	
	 if (print7)
	  System.out.println("\nnewSubTop f = " + newSubTop);
    }	 
 
 
 

 
 if (print7)
  {
	 System.out.println("\n Topology = " + topology
			 + " \nCounter of SubTops = " + counterSubtop 
	+ " \ni = " + i 
	+ "\nChosen subtopology  =  "   + subTop
	+ "\nleavesOfSubTop = " + leavesOfSubTop
	+ "\nnodesOfSubTop = " + nodesOfSubTop
	+ "\n NewSubTopology = " + newSubTop
	+ "\n New topology = " + newTopology);	
  }
 
 
 
 //System.exit(0);
 return newTopology;
}



//Mutation:
//some leaves are deleted and then reinserted 
//in random order.
//Deleting can happen in either of two ways:
//Reversing the order of synthesis else
//in random order.
//Swapping of leaves is incorporated
//and also restructuring 
//of the complement of subtopologies
public static String mutationOfTop(String topology, 
		                       int nAdditionalLeaves)
{
 original = topology; 

 if (print7)
	{
		System.out.println("\nMethod mutationOfTop" 
	+ "\nTopology = " + topology
	+ "\nNumber of right par = " + nRightParentheses
	+ "\nPresentLeaves =  "
		                   + presentLeaves);	
	}

 //Seed leaves are not removed	
 deletableLeaves = presentLeaves.substring(3);
 //Number of leaves to be deleted and then reinserted
 if (nAdditionalLeaves == 1) nUndoRedo = 1;
  else
   nUndoRedo = r.nextInt(nAdditionalLeaves-1) + 1;
 if (print1)
  System.out.println("\nNumber of leaves to delete =  "
                 + nUndoRedo);

 double s = r.nextDouble();
 if ( s < pReverseSynth)//Choose deletion in reversed order
  {
   //Some leaves are deleted: 
   //reversed order to that of synthesis.
   if (print1)
   System.out.println("Leaves are deleted " +
		                        "in reverse order");
   for(int i = 0; i < nUndoRedo; i++ )	 
	 {
	  //Last inserted leaf is deleted.
	  topology = deleteLastLeaf(topology);
	  System.out.println(topology);
	 }
 //deleted leaves are reinserted
 if (print1)
  System.out.println("\nDeleted leaves are reinserted");

 for(int i = 0; i < nUndoRedo; i++ )	 
  {	 
	  topology = generateChild(topology);
  } 
 }
 else //Choose  at random leaves to be deleted + reinserted
 topology = mutationRandomDeletion(topology, nUndoRedo);
 //Output
 String mutant = topology;
 if (print6)
  System.out.println("\nOriginal and mutant topologies = \n" 
		  + original + "\n" + mutant);
 if (print1)
  {
   if (original.equals(mutant))
	 System.out.println("Mutation made nothing");
	   else
		System.out.println("\nMutation worked");	
  }

 double d = r.nextDouble();
 if (d < 0.5)  mutant = swapLeaves(topology);

 d = r.nextDouble();
 if (d < 0.5) mutant = topComplementReestructuring(mutant);

 return mutant;
}


//Random lengths are added to a tree in Link notation 
public static void generateLengths()
{
 //Lengths with digit values are generated
 for(int i = 0; i < numberOfLeaves; i++)
  {
	 //Zero lengths are not allowed 
	 lengthsLeaves[i] = 1 + r.nextInt(maxLength - 1); 
	 lengthsInnerLinks[i] = 1 +  r.nextInt(maxLength - 1);
  }
} 



//Random lengths are added to a topology in Link notation 
public static String insertLengths(String topologyLinkN)
{
 String thisChar = "";
 String newTopology = "";
 //length of String   topologyLinkN
 int m = topologyLinkN.length();
 //topologyLinkN is scanned char by char
 for(int i = 0; i < m; i++ )
  {	 
   thisChar = "" + topologyLinkN.charAt(i); 
   newTopology = newTopology + thisChar;
   if (print1)
   System.out.println(newTopology.length() + " " 
                                 + newTopology);
   //Lengths are appended
   if (leaves.contains(thisChar))
    {	
     int k = leaves.indexOf(thisChar);
     double j = lengthsLeaves[k];
     newTopology = newTopology + ":" + j;  	
    }
 if (internalNodes.contains(thisChar))
  {	
   int k = internalNodes.indexOf(thisChar);
   double j = lengthsInnerLinks[k];
   newTopology = newTopology + ":" + j;  
  }
 }
 return newTopology;	  
}


//Lengths are deleted from newTree
private static String deleteLengths(String newTree)
{
 String tree = "";
 String numeric = ".:0123456789";
 //length of string newTree
 int l = newTree.length();
 for(int i = 0; i < l; i++ )
 {
  //newTree is scanned char by char	 
  char c = newTree.charAt(i);
  String s = "" + c;
  //Is c numeric: delete it!
  if (numeric.contains(s)); //add char c to output
	  else tree = tree + c;
 }
 return tree;	 
}


//Input: string tree with complete information.
//Work: lengthsLeaves[] and lengthsInnerLinks[]
//are read from tree.
private static void decoderLengths(String tree)
{
 if(print7)	
  System.out.println(
	"Input tree at decoderLengths = \n "   + tree  );
 //Delete :
 tree = tree.replaceAll(":", "");
 if(print7)	
 System.out.println(
			"Input tree without : = \n "   + tree  );
 int n = tree.length();
 String thisChar = "";
 String nextChar = "";
 String numbers = ".0123456789";
 boolean leaf = false;
 boolean node = false;
 String Leaf = "";
 String numberLeaf = "";
 String numberNode = "";
 String Node = "";
 int j ;
 double f;

//All chars in tree are scanned
for(int i=0; i < n; i++)
{
 thisChar = "" + tree.charAt(i);
 if (i < n-1) nextChar  = "" + tree.charAt(i+1);
 //Do we have a leaf?
 if (leaves.contains(thisChar)) 
   {
	 leaf = true; 
	 Leaf = "" + thisChar;
   }

 //Do we have an internal node?
 if (internalNodes.contains(thisChar)) 
  {
	 node = true;  
	 Node = "" + thisChar;
  }

 //Do we have a number that belongs in a leaf?
 if ( (leaf) & (numbers.contains(thisChar)))
	{
	 numberLeaf = numberLeaf + thisChar; 
	}

 //Do we have a number that belongs in a leaf?
 if ( (node) & (numbers.contains(thisChar)))
	{
	 numberNode = numberNode + thisChar; 
	}

 //Must we end accumulation of numbers of a leaf?
 if ( (leaf) & (internalNodes.contains(nextChar)))
	{
	  leaf = false;
	  j = leaves.indexOf(Leaf);
	  if (print1)
	  System.out.println("Leaf = " + Leaf +
			  " numberLeaf = " + numberLeaf);
	  f = new Double(numberLeaf);
	  lengthsLeaves[j] = (int) Math.floor(f);
	  numberLeaf = "";
	}

 //Must we end accumulation of numbers of a node?
 if ( (node) & (!(numbers.contains(nextChar))))
	{
	  node = false;
	  j = internalNodes.indexOf(Node);
	  if (print1)
	  System.out.println("Node = " + Node);
	  f = new Double(numberNode);
	  lengthsInnerLinks[j] = (int) Math.floor(f);
	  numberNode = "";
	}

 } 	 
}


//The length of a link can decrease or increase
//by one
private static  void tinyMutations()
{
 //Mutation in length of links that connect leaves
 for(int k = 0; k < numberOfLeaves; k++)
   {
	double q = r.nextDouble();		
	if (q < m) //mutate
	   {   
		   double f = r.nextDouble();	   
	   if (f < m) //mutate
			lengthsLeaves[k] = lengthsLeaves[k] + 1 ;	
	   else
		   lengthsLeaves[k] = lengthsLeaves[k] - 1 ;
	   
	   if (lengthsLeaves[k] > 9) 	lengthsLeaves[k] = 9;
	   if (lengthsLeaves[k] < 0) 	lengthsLeaves[k] = 0;
	   }
   }
  //Mutation in length of links that connect inner nodes
  for(int k = 0; k < numberOfLeaves-2; k++)
   {
	double q = r.nextDouble();
	if (q < m) //mutate
	 {   
	  double f = r.nextDouble();	   
	  if (f < m) //mutate
		lengthsInnerLinks[k] = lengthsInnerLinks[k] + 1 ;	
	   else
		lengthsInnerLinks[k] = lengthsInnerLinks[k] - 1 ;
	   
	  if (lengthsInnerLinks[k] > 9) lengthsInnerLinks[k] = 9;
	  if (lengthsInnerLinks[k] < 0) lengthsInnerLinks[k] = 0;
	}
  } 	
}


//Correlated mutations in the length of links
//of the form (+/-1, +/-1). 
//The first change is made in the length associated 
//to a leaf and the second in that of the accompanying node.
private static  String tiny2Mutations(String tree)
{
 //A leaf is chosen at random	
 int n = presentLeaves.length();
 int i = r.nextInt(n);
 String leaf = "" + presentLeaves.charAt(i);
 int j = tree.indexOf(leaf);
 String oldString = tree.substring(j, j+10);
 String node = "" + tree.charAt(j+5);

 char t1 = tree.charAt(j+2);
 char t2 = tree.charAt(j+7);
 
 Integer k1 = Character.getNumericValue(t1);
 Integer k2 = Character.getNumericValue(t2);
 double g = r.nextDouble();
 if (g < 0.5)
  {
   if ( k1 < 9) k1 = k1 +1;
   if (k2 > 0) k2 = k2-1;
  }
  else
   {
	 if ( k1 < 9)  k2 = k2 +1;
	 if (k2 > 0)  k1 = k1-1; 
   }
 
 String newString = leaf + ":" + k1.toString() + ".0"
		  + node + ":" + k2.toString() + ".0";
 
 String output = tree.replace(oldString, newString);
 if (print1)
	  System.out.println("Tree   = " + tree 
			             +  "\nOldString = " + oldString
			             + " leaf = " + leaf
			             + " node = " + node
			             + " t1 = " + t1
			             + " t2 = " + t2
			             + " k1 = " + k1
			             + " k2 = " + k2
			             + " newString = " + newString
			             + "\nOutput = " + output); 
 	
 //Give just one length to each node and leaf
 correction();
 return output;

}




//Lengths undergo mutation:
//Choose tiny mutation (+/-1) else
//the old  procedure.
private static  void mutateLengths(String tree)
{	
 double f = r.nextDouble();
 if ( f < pTinyMutations ) //new procedure
  {
	tinyMutations();
	tiny2Mutations(tree);
  }
  else //Old procedure
   {

     //Mutation in length of links that connect leaves
     for(int k = 0; k < numberOfLeaves; k++)
      {
	   double q = r.nextDouble();		
	   if (q < m) //mutate
			lengthsLeaves[k] = r.nextInt(maxLength);		 	
      }
    //Mutation in length of links that connect inner nodes
    for(int k = 0; k < numberOfLeaves-2; k++)
     {
	  double q = r.nextDouble();
	  if (q < m) //mutate
			lengthsInnerLinks[k] = r.nextInt(maxLength);		 	
     }
   }
}



//Type of  mutation is decided and executed:
//mutation of lengths of links
//else full mutation (topology and lengths of links).
private static String mutateTree(String tree)
{ 
 presentLeaves = leavesOfTopol;
 //Topology is read from tree
 String topology = deleteLengths(tree);
 //Lengths are read from tree and recorded in vectors
 //lengthsLeaves[] and lengthsInnerLinks[]
 decoderLengths(tree);
 mutateLengths(tree);
 String  mutantTree = "";
 //Mutate lengths else topology ( + mutated lengths)?
 double s = r.nextDouble();
 if ( s < pMutateOnlyLengths) //only lengths are mutated
	{
	 if (print1)
	 System.out.println("\n====Mutation of lengths");
	  
	  mutantTree = 	insertLengths(topology);

	  if (print1)
	  {
	  System.out.println("Input tree =  " + tree);
	  System.out.println("Topology =  " + topology);
	  System.out.println("Output tree= " + mutantTree);
	  }
	 }
 else //topology + lengths are mutated
	{
	 if (print1)
	  System.out.println("\n====Mutation of topology");
	 
 
   //Mutation of topology
	 String mutantTopology = mutationOfTop(topology, 
			                           nAdditionalLeaves);

	 //Mutant lengths are reinserted
	 mutantTree = 	insertLengths(mutantTopology);
	 if (isTree(mutantTree));
	 /*
	 else
	  {
		System.out.println("New tree is not  tree " +
				"at mutateTree = \n" + mutantTree);
		System.exit(0);
	  }
	  */
	}
	
 return mutantTree;
}





//Individuals are mutated
private static void mutation()
{
 if (print5)
  {
	System.out.println("\n====Mutation \n");
	
	System.out.println("Trees to be mutated");
  }
 for(int i = 0; i < N; i++)
	 {
	  trees[i] = newTrees[i];
	  if (print5)
	  System.out.println(trees[i]);
	 }
 //The old best, i = 0, is conserved intact
 newTrees[0] = trees[0];
 //Other N-1 individuals suffer mutation
 for(int i = 1; i < N; i++)
  newTrees[i] = mutateTree(trees[i]);


if (print5)
	System.out.println("Mutant trees");
	 for(int i = 0; i < N; i++)
	 {
	  trees[i] = newTrees[i];
	  if (print5)
	  System.out.println(trees[i]);
	 }
}

//==========Recombination============


 //Counter of subtopologies
 private static  int counterSubtop;
 //Particular number of subtopologies
 private static  int m1 = 0;



 private static  String[] subtrees1, subtrees2;
 private static  String   subTree1, subTree2,
                     treeRec1, treeRec2;

 //Structure of hashtable : integer + {string1, string2}
 private static Hashtable<Integer, String[]> hashtable =
		 new  Hashtable<Integer, String[]>();

 

//========== Subtrees are listed ============


//Subtopologies of  topology are found
private static String[] subtopologies(String topology)
{
 if (print5)
  System.out.println("\nMethod subtopologies(string)");
 if (print5)
	  System.out.println("\nTopology = " + topology);
 //Counter of left parenthesis
 int counter = 0;
 //Counter of subtopologies
 counterSubtop = 0;
 String pl = "" + '(';
 String pr = "" + ')';
 //Length of String   topology
 int l = topology.length();
 String[] subtopologies = new String[l];
 //String topology is scanned char by char
 for(int i = 0; i < l-1; i++ )
  {	 
   String Sub = "";
   String a = "" + topology.charAt(i);
   int position = i+1;
   if (a.equals(pl))
   {   
    counter++;
    Sub =  Sub + a;
    //The complete subtopology is captured:
    //Parentheses must balanced
   while (counter > 0)
    {   
     String b = "" + topology.charAt(position);
     if (b.equals(pl))   counter++;
     if (b.equals(pr))  counter--;
     Sub =  Sub + b;	 
     position++;
    } 	
   if (print5)
	   System.out.println(Sub);
   subtopologies[counterSubtop] = Sub; 
   counterSubtop++;
   }
  }	 
 return subtopologies;
}


//Leaves of subtopologies  are found
private static String[]leaves(int m, String[] subtop)
{
 if (print1)  
  System.out.println("Leaves are found");
 String[] myLeaves = new String[m];
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
   else
    {
     myLeaves[i] = "";
     int n = s.length();
     for(int j = 0; j < n; j++)
      {
       String c = "" + subtop[i].charAt(j);
       if (leaves.contains(c))
 	     myLeaves[i] = myLeaves[i] + c;
     }	
   if (print1)
     System.out.println(subtop[i] + " "
                        + myLeaves[i] );
   }	
  }
 //Chars of Leaves are rewritten in alphabet order
 if (print1)
  System.out.println("Leaves are listed orderly" );
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
    else
     {
      s =   myLeaves[i];
      myLeaves[i] = "";
      int n = leaves.length();
   for(int j = 0; j < n; j++)
    {
     String c = "" + leaves.charAt(j);
     if (s.contains(c))
 	   myLeaves[i] = myLeaves[i] + c;
    }	
   if (print1)
   System.out.println(subtop[i] + " " + myLeaves[i] );
  }	
 } 
 return myLeaves;
}



//Inner nodes of subtopologies  are found
private static String[] inner(int m, String[] subtop)
{
 if (print1)
  System.out.println("\nInternal nodes are found");
 String[] myNodes = new String[m];
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
    else
     {
      myNodes[i] = "";
      int n = s.length();
      for(int j = 0; j < n; j++)
       {
        String c = "" + subtop[i].charAt(j);
        if (internalNodes.contains(c))
  	     myNodes[i] = myNodes[i] + c;
       }	
    if (print1)
	System.out.println(subtop[i] + " " + myNodes[i] );
  }	
 }
 //Chars of Leaves are rewritten in alphabet order
 if (print1)
 System.out.println("Inner nodes are listed orderly" );
 for(int i = 0; i < m; i++)
  {
   String s = subtop[i];
   if (s == null);
    else
     {
      s =   myNodes[i];
      myNodes[i] = "";
      int n = internalNodes.length();
      for(int j = 0; j < n; j++)
       {
        String c = "" + internalNodes.charAt(j);
        if (s.contains(c))
    	 myNodes[i] = myNodes[i] + c;
       }
  if (print1)
   System.out.println(subtop[i] + " " + myNodes[i] );
 }	
} 
return myNodes;
}


//Subtopologies are compared to decide whether
//or not they have the same set of leaves
//and the same set of inner nodes.
//All pairs that authorize recombination
// are kept in a hastable (=  a table for pairing)
private static boolean comparison(String Leaves1[], 
	                       String Leaves2[],   
	                       String Inner1[], 
	                       String Inner2[]) 
{
 if (print5)
 System.out.println("\nComparison of leaves and nodes");
 Boolean recAllowed = false;
 int m = Leaves1.length;
 int n = Leaves2.length;
 Integer counter = 0;
 for(int i = 1; i < m; i++)
  {
   for(int j = 1; j < n; j++) 
    {
     if ( Leaves1[i].equals(Leaves2[j])) 
      {
       subTree1 = subtrees1[i];	
       subTree2 = subtrees2[j];	
       recAllowed = true;
       if (print5)
        System.out.println(
          "Sets of leaves are EQUAL " + i + 
        "  " + Leaves1[i] + "  " + j + "  " + Leaves2[j] );
        String val[] = {subTree1, subTree2};

        //Recombinant subtrees are kept in a hashtable
        hashtable.put(counter,val);
        counter++;
        if ( Inner1[i].equals(Inner2[j]))   
         {
	      if (print5)
	 	    System.out.println(
	        "Sets of inner nodes are EQUAL " + i +
          "  " + Inner1[i] + "  " + j + "  " + Inner2[j] );
         }
        else 
         {
	       if (print5)
		   System.out.println(
	       "Sets of inner nodes are DIFFERENT " + i +
	       "  " + Inner1[i] + "  " + j + "  " + Inner2[j] );	 
          }
      }
    else 
	  if (print5)
	   System.out.println(
       "Sets of leaves are DIFFERENT " + i +
        "  " + Leaves1[i] + "  " + j + "  " + Leaves2[j] );	
   }
 }	
 return recAllowed;
}


//A pair of subtrees that authorize recombination
//is chosen at random from the hashtable and 
//recombination is executed.
private static  String recombinationTrees(String tree1, 
                                          String tree2)
{
 //Possible pairs of recombinant subtrees are recovered
 //from the hashtable and put into an enumeration<>
 Enumeration<Integer> enumKey = hashtable.keys();
 int counter = 0;
 while(enumKey.hasMoreElements())
  {
   Integer key = enumKey.nextElement();
   String[] val = hashtable.get(key);
   if (print3)
    System.out.println("Hasthtable " + key    + "\n"
		 + val[0]  + "\n"+  val[1]);
   counter++;
   }

 //A pair of recombinant subtrees is chosen at random
 Integer election = r.nextInt(counter);
 if (print5)
  System.out.println("Selected pair = " + election);
 String[] val = hashtable.get(election);
 subTree1 = val[0];
 subTree2 = val[1];
 if (print5)
   System.out.println("Recombinant subtrees \n" +
      subTree1 + "\n" +  subTree2);

 //Recombination is executed
 treeRec1 = tree1.replace(subTree1, subTree2);
 treeRec2 = tree2.replace(subTree2, subTree1);
 if (print5)
  System.out.println("\nRecombination is allowed:"
	 + "\nOriginal trees \n"
	 + tree1 + "\n" + tree2
	 + "\nSwapping subtrees \n" 
	 +  subTree1 + "\n" + subTree2
	 + "\nRecombinant trees \n"
	 + treeRec1 + "\n" + treeRec2 );
 return treeRec1;
}


//Recombination of tree1 and tree2 is tried out
public static String recombinationMaster(String tree1, 
		                              String tree2) 
{   
 //First topology  
 String topo1 = deleteLengths(tree1);
 if (print5)
  {
   System.out.println(topo1 + " -> " 
     + linkToNewick(topo1));
   System.out.println("\n Tree1 = " + tree1 +
			 "\n Topo1 = " + topo1);
  }

 //Length of String   topology
 m1 = topo1.length();

 //Subtopologies are found
 subtrees1 = new String[m1];
 subtrees1 = subtopologies(topo1);
 m1 = counterSubtop;
 String[] subtopologies1 = new String[m1];
 for(int i = 0; i < m1; i++)
 subtopologies1[i] =  simplify(subtrees1[i]);
 if (print5)
  System.out.println();
 String[] Leaves1 = new String[m1];
 Leaves1 = leaves(m1,subtopologies1);
 String[] Inner1 = new String[m1];
 Inner1 = inner(m1,subtrees1);
 //Chars of Leaves are rewritten in alphabet order
 if (print5)
  System.out.println("\nSubtrees, subtopologies, Leaves, " +
	             "Inner nodes" );
 for(int i = 0; i < m1; i++)
  {
   if (subtopologies1[i] == null);
    else
    {   
	 if (print5)  
      System.out.println(i 
		   + "   " + subtrees1[i]
				   + "   " + subtopologies1[i]
		   + "   " +  Leaves1[i] + "   " + Inner1[i]);
    }
  }

 if (print5)
  System.out.println();

 //Second topology 
 String topo2 = deleteLengths(tree2);
 if (print5)
  {
   System.out.println(topo2 + " -> " 
    + linkToNewick(topo2));

	 System.out.println("\n Tree2 = " + tree2 +
			 "\n Topo2 = " + topo2);	
  }
 //Length of String   topology
 int  m2 = topo2.length();

 subtrees2 = new String[m2];
 subtrees2 = subtopologies(topo2);
 m2 = counterSubtop;
 //Subtopologies are found
 String[] subtopologies2 = new String[m2];
 for(int i = 0; i < m2; i++)
  subtopologies2[i] =  simplify(subtrees2[i]);


 if (print5)
  System.out.println();
 String[] Leaves2 = new String[m2];
 Leaves2 = leaves(m2,subtopologies2);
 String[] Inner2 = new String[m2];
 Inner2 = inner(m2,subtrees2);
 //Chars of Leaves are rewritten in alphabet order
 if (print5)
  System.out.println("\nSubtopologies, Leaves, " +
	             "Inner nodes" );
 for(int i = 0; i < m2; i++)
  {
   if (subtopologies2[i] == null);
   else
    {   
	 if (print5)
     System.out.println(i 
		   + "   " + subtrees2[i]
				   + "   " + subtopologies2[i]
		   + "   " +  Leaves2[i] + "   " + Inner2[i]);
    }
   }   
 boolean recAllowed = 
	   comparison(Leaves1,Leaves2, Inner1,Inner2); 
 String recombinant = "";
 if (recAllowed)
  {
   recombinant = recombinationTrees(tree1,tree2);
   if (print5)
		  System.out.println("Recombinant tree = "
				  + recombinant );   
   }

  else 
   {
    recombinant = tree1;
	
    if (print5)
	  System.out.println("Recombination is not allowed");
   }
 return recombinant;
}


//Recombination methods may produce an output
//with various and different assignment of lengths to 
//inner node a. One is chosen.
public static void correction() 
{ 
 String topo;
 for(int i = 1; i < N; i++)
  {
	decoderLengths(newTrees[i]);
	topo = deleteLengths(newTrees[i]);
	newTrees[i] = insertLengths(topo);
  }
}

//Each one of the N individuals  are tried out
//for recombination: a partner is chosen at random
//and if recombination is allowed, it is executed.
//Otherwise, it is cloned.
public static void recombination() 
{ 
 if (print5)	
 System.out.println( "\n====RECOMBINATION \n");
 //Best individual is excluded from recombination
  newTrees[0] = trees[0];

 for(int i = 1; i < N; i++)
  {
   //A partner for i is chosen	 
   int j = i;
   String t1 = "";
   String t2 = "";
   while (i == j)
	 j =   r.nextInt(N);
   if (print5)	
	System.out.println( "\n==RECOMBINATION OF INDIVIDUALS "
			+ i + " and " + j);   
   t1 = trees[i];
   t2 = trees[j];   
   newTrees[i] = recombinationMaster(t1,t2);
   correction();
  }
 if (print5)
 System.out.println("Recombination output");
 for(int i = 0; i < N; i++)
 {
  trees[i] = newTrees[i];
  if (print5)
  System.out.println(trees[i]);
 }
}



//Evolution of trees involving
//recombination and mutation of topologies,
//in within and between.
private static void evolution()
{
 for(int i = 1; 
	 ((i < numberOfGenerations) & (workDone)); i++)
	
 {
	generation = i;
  if ( (i % lapse) == 0) 
		  {
	              print8 = true;
	          //  a.setEnabled(false);  
		  }
  else print8 = false;
  
  if (print8)
  System.out.println( "\n===GENERATION  = " + i);
  selectiveCloning();
  mutation();
  recombination();
  assessFitnesses();
  scaleFitnesses();
 
 }
 
 
 
 bestTopology = simplify(deleteLengths(bestTree));
 
 title1 = "Target tree = "  + targetTree;   
 title2 = "Best     tree = "   + bestTree; 
 title3 = "Error = Distance from matrix of " 
 		  + "best fitting tree" 
 		 + " to  target matrix = " + bestDistance;
 
 decoderLengths(bestTree);

 //Absolute distance from the associated distance matrix
 //of the tree
 //to target matrix is evaluated
 //d = absDistance(numberOfLeaves,targetMatrix, M);
 
 
 
 
 double[][] M = associatedMatrix(bestTree); 
 
 draw = newContentPane.new DrawBiTree18();
 draw.createAndShowGUI();
 draw.setLocation(500,  100); 
 
	
 if (!(workDone))
  System.out.println( "\n \nSOLUTION HAS BEEN FOUND: \n" +
			"at generation " + generation);
 
 System.out.println("\nSummary at  generation " + generation + 
		 "\nTarget tree   = \n" + targetTree		
		   + " \nBest tree = \n" + bestTree
		   + " \nDistance to target = " + bestDistance
		   + " \nBest distance at initialization = " 
		   + distance0
		   + "\nTarget topology = \n" + targetTopology
		   + "\nBest topology = \n" + bestTopology);
 printMatrix(M, "Distance matrix associated to best tree");
 printMatrix(targetMatrix, 
	     	"Target distance matrix");
}



//The best fitting tree for chosen topology is found
public static int findTheBest(String tree)
{
 String selectedTopology = simplify(tree);	
 String thisTopology = "";
 String thisTree = "";
 int bestPosition = -1;

 for(int i = 0; i < indexOfTopologies; i++)
  {
   thisTree = bestTreeOfTopology[i];
   thisTopology = simplify(thisTree);
   if (selectedTopology.equals(thisTopology))
    {
     bestPosition = i;
     return bestPosition;	    
    }
  }
 return  bestPosition;
}


  /*
  //Lengths with real values are generated
  double lengthsLeaves[]  = new double[l];
  //Last two entries are redundant
  double lengthsInnerLinks[]  = new double[l];  
  for(int i = 0; i < numberOfLeaves; i++)
   {
	 lengthsLeaves[i] = r.nextDouble(); 
	 lengthsInnerLinks[i] =  r.nextDouble();
   }
  */  

  
 
  
 




//========= Translation in within formats ==========
  
  

//bTree in (modified) Newick's notation 
//is used to generate Record[]
//with addresses and contents.
public static void makeInput(String bTree) 
{
 if (print)
 System.out.println("bTree at makeInput ////// " + bTree);
 int counter = 0;
 //length of String   tree
 int l = bTree.length();
 String thisChar = "";
 String address = "";
 String binNumbers = "01";
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + bTree.charAt(i);
	if (binNumbers.contains(thisChar)) 
		   address = address + thisChar;
	if (leaves.contains(thisChar)) 
	{
	 Record[counter] = address;
     counter++;
	 Record[counter] = thisChar;
	
	counter++;
	address = "";
   }
  }
 
 if (print)
  for(int i = 0; i < counter; i = i+2 )
	  System.out.println(Record[i] +
			 " bbbb " + Record[i+1]);

 lastUsedPosition = counter;
 if (print)
	System.out.println("lastUsedPosition ==== " 
        + lastUsedPosition); 
}



//Outputs a word in reversed order.
private static String reverser(String word)
{
 return new StringBuilder(word).reverse().toString();	
}


private static String cleanTree(String tree)
{
 if (print2)
 System.out.println("Input tree at cleanTree = \n "
        + tree);
 String treeClean = "";	
 String numeric = ".:0123456789";
 String plusMinus = "+-";
 String thisChar = "";
  boolean leaf = false;
 //Numerics are washed
 //length of String   tree
 int l = tree.length();
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + tree.charAt(i);
   if (numeric.contains(thisChar));
    else treeClean = treeClean + thisChar;	  
  }
 if (print2)
 System.out.println("Tree without numerics = \n"
		+ treeClean);
 //+/- are deleted from inner nodes that accompany leaves
 l = treeClean.length();
 String treeOut = "";
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
 {	 
  thisChar = "" + treeClean.charAt(i);
  if (leaves.contains(thisChar))
	  leaf  = true;
  if ((leaf) &  (plusMinus.contains(thisChar)));
	else treeOut = treeOut + thisChar;	  
  if (internalNodes.contains(thisChar))
	  leaf  = false;
}

 if (print2)
 System.out.println("Tree at output of cleanTree = \n"
			+ treeOut);
return treeOut;

}




//The path from each leaf to node a is found.
//The input contains branch's indicator, say:
//((++Aa,+-Ba),(-+Cb,((--++Dd,--+-Fd)--c,---Ec)-b)+a)
//Output for leaf A: Aa+.
//Output for leaf C: Cba-
//Output for leaf D: Ddcba-
private static void findPaths(String tree)
{
 if (print2)	
 System.out.println("Input tree at findPaths = \n" + tree); 
 //length of String   tree
 int l = tree.length();
 String plusMinus = "+-";
 String thisChar;
 String nextChar;
 String path = "";
 //Branch indicative with respect to root, + else -
 String branch;
 String address ;
 //The number of links of a path to root
 int level = 0;
 String treeReversed = reverser(tree);
 if (print2)
 System.out.println("Reversed tree = " + treeReversed);
 
//Each leaf has a path to a	 
for(int k= 0; k < numberOfLeaves; k++)	 
{
 String leaf = "" + leaves.charAt(k); 
 //Candidate nodes to be part of the path to node a
 //must have  sign +/- as prefix.
 //Or as suffix in the reversed tree.
 int position = treeReversed.indexOf(leaf);
 if (position > -1) // if leaf is in tree
 {
  if (print2)
  System.out.println("\nLeaf = " + leaf ); 
  address = "";
  boolean go = true;
  //branchIndicator and level are found
  for(int i = position+1 ; ((i < l) & (go)); i++ )
  {	 
   thisChar = "" + treeReversed.charAt(i);  
   if   (plusMinus.contains(thisChar)  )
	address = thisChar + address;
   else go = false;   
  }  
  level = address.length();
  if (print2)
  System.out.println("LeafAddressLevel  = " + leaf 
		                     + address + level); 
  branch = "" + address.charAt(0);
 
  
  //Path from leaf to node a is found in (not reversed) tree 
  position = tree.indexOf(leaf);
  thisChar = "" + tree.charAt(position); 
  nextChar = "" + tree.charAt(position + 1);
  //Output
  path = "" + leaf + nextChar;
  go = true;
  if  (nextChar.contentEquals("a") ) //leaf is neighbor of a
   {
	  path = path + branch;
	  if (print)
	  System.out.println("Path short = " + path);
   }
  else	 //level > 1
  {	 	  
	  
  //Tree is scanned char by char to get path
  String iChar = "";
  for(int i = position + 2; ((i < l) & (go)); i++ )
  {	 
   iChar = "" + tree.charAt(i);
   
  
   int newLevel = 0;
   if (internalNodes.contains(iChar))//InternalNodes
	{	   
     if   (iChar.contentEquals("a")  )
      {
	   path  = path +  iChar;
	   go = false; //end
      }
     else //Internal node different than a
     {
      if (print)	 
      System.out.println("Internal node not a) = " + iChar );
      int rP = l -i-1;	//index of iChar in treeReversed       
      String  nextNewChar = "" + treeReversed.charAt(rP+1);
      //candidate node
      if ( plusMinus.contains(nextNewChar )) 
       {  
    	  if (print)  
    	System.out.println("next candidate = " + iChar); 
        String addressNew = "";
        
        //level of candidate node iChar  is inquired
        String thatChar = "";
        boolean go2 = true;
        for(int j = rP+1 ; ((j < l) & (go2)); j++ )
         {	 
    	  thatChar = "" + treeReversed.charAt(j);  
    	  if   (plusMinus.contains(thatChar)  )
          addressNew = thatChar + addressNew;
    	  else go2 = false;
         }  
        if (print)
        System.out.println("AddressNew " +
                                    iChar + addressNew);
       newLevel = addressNew.length();
       //A path links a node with another in a lower level
       if  ( (newLevel < level) &
    		   (!(nextChar.contentEquals(iChar))))
        {
    	  path = path + iChar;
    	  level = newLevel;
    	  if (print)
    	  System.out.println("Path newLevel= " + path);
        }
      } //End of if candidate node    	 	 
     }//End of else Internal node different than a 
	}//End of if InternalNodes 
   }//End of Tree is scanned char by char
  
  path = path + branch;
  
 }//End of  else level > 1    
 }//End of if leaf is in tree
 if (print2)
 System.out.println("Path completed  = " + path);
 Paths[k] = path;
}//End of each leaf
}



//Paths from leaves to node a are found
//Reverse paths are also calculated.
//Input tree in quasibinary notation 
//with addresses and lengths
private static void pathFinder(String tree)	
{
 //Useless information is deleted	
 String tree2 = cleanTree(tree);
 //Find the path to node a from each leaf.
 //Add an indicator of the branch
 //with respect to the root: up(+) else down(-)
  findPaths(tree2);  	
}

//Input: a long raw string with path from leaf to node a
//plus path from node a to leaf.
//Editing is done to get true paths.
private static String grammar (String path)
{
 int n = path.length();
 String thisChar = "";
 String nextChar = "";
 String oldPath = path;
 String plusMinus = "a+-a";
 String minusPlus = "a-+a";
 String two = "";
 int N = path.length();
//Rule 1: two equal symbols produce nothing
 for(int j=1; j < N; j++)
  {
   for(int i=1; i < n-1;)
   {	 
    thisChar = "" + path.charAt(i);
    nextChar = "" + path.charAt(i+1);
    two = "" + thisChar + nextChar;
    if (thisChar.contentEquals(nextChar)) 
	  {
	   i = i + 2; 
	   path = path.replace(two, "");
	  }
    else 
	  {	   
	   i = i+1;
	  } 
  n = path.length();
   } 
  }
 
 //Rule 2: a+-a -> a
 
 if  (path.contains(plusMinus))		
      path = path.replace(plusMinus, "a");
 if   (path.contains(minusPlus))
     path = path.replace(minusPlus, "a");
 if (print2)   
 System.out.println("input = " + oldPath + 
		                    " output = " + path);
 return path;	
}

//The distance matrix among leaves is calculated
private static double[][] matrixBuilder()
{
	double M[][] = 
	new double[numberOfLeaves][numberOfLeaves];
 //Path to node a from leaf
 String pathLeafToa = "";
 //Path to leaf from node a
 String pathaToLeaf = "";
 //Path from leaf to lead
 String pathLeafiToLeafj = "";
 String innerNode = "";
 double distance;
 for(int i = 0; i < numberOfLeaves; i++)
  for(int j = 0; j < numberOfLeaves; j++)
  {
   pathLeafToa = Paths[i];
   pathaToLeaf = reverser( Paths[j] );
   //Raw paths
   pathLeafiToLeafj = pathLeafToa + pathaToLeaf;
   //An editing grammar to get true paths
   pathLeafiToLeafj = grammar( pathLeafiToLeafj );
   if (i == j) M[i][j] = 0;
   else
    {	 
	 String leaf = "" + pathLeafiToLeafj.charAt(0);
	 int p = leaves.indexOf(leaf);	 
	 distance = lengthsLeaves[p];
	 int n = pathLeafiToLeafj.length();
	 for(int k = 1; k < n-1; k++)
	 {	 
	  innerNode =  "" + pathLeafiToLeafj.charAt(k);
	  p = internalNodes.indexOf(innerNode);	 
	  distance = distance + lengthsInnerLinks[p];
	 }
	 leaf =  "" + pathLeafiToLeafj.charAt(n-1);
	 p = leaves.indexOf(leaf);	 
	 distance = distance + lengthsLeaves[p];
	  M[i][j] = distance;
    }
  }
 return  M;
}



//Translation from Newick's to quasi-binary notation:
//addresses are attached to tree in Newick's notation
//Inner nodes are conserved. Lengths are conserved.
private static String toQuasiBinaryInner(String tree) 
{
 if (print2)
  System.out.println("Input tree at toQuasiBinaryInner " 
		                   +    "= \n" + tree); 	
 //Output
 String treeToBi = "";	
 String numeric = ".:0123456789";
 //The binary address of leaf
 String address = "";
 //The name or content of the tree
 String content = "";
 String thisChar = "";
 String nextChar = "";
 String previousChar = "";
 //length of String   tree
 int l = tree.length();
 //Tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + tree.charAt(i);
   if ( i < l-1) nextChar = "" + tree.charAt(i+1);
    else nextChar = "end";
   if ( i > 0) previousChar = "" + tree.charAt(i-1);
    else previousChar = "start";


   //This char = (
   if (thisChar.contentEquals("(")) 
    {
     content = content + thisChar;
     address = address + '+';
     if (print)
	 System.out.println("Address = " + address); 		
    }
	   
	   
 //If this char is a leaf 
 if (leaves.contains(thisChar))
  {
   // System.out.println("leaf"); 
   treeToBi =  treeToBi + content + address + thisChar;
   content = "";
  }

 if (numeric.contains(thisChar))
	 treeToBi =  treeToBi + thisChar;
	 
 //If this char is an inner node
 if (internalNodes.contains(thisChar))
  {
   int m = address.length();  
			//delete last bit from the right
   if (m>0)	address = address.substring(0, m-1) ;
   if (address.contentEquals("")) address = "+";
   treeToBi =  treeToBi +	address + thisChar;	 		
   
  }
	 
	 
 //This char = , and goes after an inner node
 if  (thisChar.contentEquals(",")) 
  {
   //System.out.println
	 //("comma : "  + previousChar + thisChar ); 
   treeToBi =  treeToBi +	thisChar ;
   if (print)
	 System.out.println("Address at ,1   = " + address); 
		   
   if    (numeric.contains(previousChar))
		  address = address + '-';//Add a - to the right
   if (print)
	 System.out.println("Address at ,2   = " + address);
  } 
	   
	   
 if (thisChar.contentEquals(")")) 
  {
	treeToBi =  treeToBi +	thisChar ;
    //Diphthonge   ),
	if   (nextChar.contentEquals(","))
	{ 
     int m = address.length(); 
	 //delete last bit from the right  
	 if (m >0)   address = address.substring(0, m-1) + "-";
   }
  }
	  		
}//End of for(int i...
return treeToBi;	
}


//bTree  (in links with lengths notation) 
//is used to generate Record[]
//with addresses and contents.
//Names of inner nodes are visible
public static void makeInputInner(String bTree) 
{
 int counter = 0;
 //length of String   tree
 int l = bTree.length();
 String thisChar = "";
 String address = "";
 String binNumbers = "+-";
 String numeric = ".:0123456789";
 //String tree is scanned char by char
 for(int i = 0; i < l; i++ )
  {	 
   thisChar = "" + bTree.charAt(i);
	if (binNumbers.contains(thisChar)) 
		   address = address + thisChar;
	if ( (leaves.contains(thisChar)) ||
			(internalNodes.contains(thisChar)))
	{
	 Record[counter] = address;
     counter++;
	 Record[counter] = thisChar;
	
	 counter++;
	 address = "";
    }
	
	if (numeric.contains(thisChar))
	 {
		Record[counter-1] = Record[counter-1] + thisChar;
	 }
	
  }


 if (print)
  for(int i = 0; i < counter; i = i+2 )
	  System.out.println(Record[i] +
			 " bbbb " + Record[i+1]);

 lastUsedPosition = counter;
 if (print)
	System.out.println("lastUsedPosition ==== " 
      + lastUsedPosition); 
}//End of method


//Prints  a matrix and its title
private static void printMatrix(double Matrix[][], 
		                        String title)
{
 System.out.println( "\n" + title );
 int m = Matrix.length;
 System.out.print(  "\t");
 for ( int j = 0; j < m; j++ )
  System.out.print( leaves.charAt(j)+ "\t");		
 System.out.println();
 for ( int i = 0; i < m; i++ )
  {
   System.out.print(leaves.charAt(i ) + "\t");	
   for ( int j = 0; j < m; j++ )
	 System.out.print( Matrix[ i ][j]+ "\t");
   System.out.println();
  }
}


//Tree is translated 
//from link notation to binary.
//Output in Record[]: even entries= addresses, odd = names.
//Associated matrix is also calculated
public static double [][] associatedMatrix(String tree) 
{
 /*	
 System.out.println("Tree  at translate = " + tree);
 //String NwTree = simplify(tree);
 String bTree = toBinary(tree);
 //A vector with addresses and names is composed
 makeInput(bTree);
 */
 //String NwTree = simplify(tree);
 String bTree = toQuasiBinaryInner(tree);	
 
 if (print2)
 {	 
 System.out.println("Input tree  at translate = "  + tree);
  System.out.println("Selected tree in " +
	 	"pseudo-Newick's notation = \n" + simplify(tree));
  System.out.println("Tree in quasi-binary notation: \n"
		               + bTree);
 }
 
 //A vector with addresses and other with names are composed
 makeInputInner(bTree);	
 //Paths from leaves to node a are found.
 //Reverse paths are also listed.
 pathFinder(bTree);	
 //The distance matrix among leaves is calculated
 double M[][] = matrixBuilder();
 return M;
}
 



//The absolute distance among two matrices is calculated.
//Only triangular part of matrices are compared. 
private static double absDistance(int n, 
		               double[][] M1, double[][] M2)
{
 double abs = 0;
 for(int i = 0; i < n; i++)
 for(int j = i+1; j < n; j++)
   abs = abs + Math.abs(M1[i][j]-M2[i][j]);  
 return abs;
}


//Input : tree in link notation.
//Output:  
//distance matrix associated to the leaves of the tree.
private static double[][] distanceMatrix(String tree)
{
 String s = toQuasiBinaryInner(tree) ;	
 if (print4)
	 System.out.println("\n Tree = " + s);	
 //Paths from leaves to node a are found.
 //Reverse paths are also listed.
 pathFinder(s);	
 //The distance matrix among leaves is calculated
 double M[][] = matrixBuilder();	
 return M;
}



//===========================================
//==Inner class  for drawing selected tree ==
//===========================================


//This program translates a string that 
//denotes a tree in Newick's notation
//to our binary notation
//(assign a zero to the left, a one to the right).

//A generator of random trees in Newick's notation
//is included.

//Resultant trees are drawn.


	


class DrawBiTree18 extends JPanel
                      
{
 private static final long serialVersionUID = 1L;
 private JPanel drawingPane38;
 JScrollPane scroller;


 

 //=======================================



 //Vertical  size of the panel 
 int verticalSize;
 int horizontalSize;
//Scale must be a power of 2
 int scale = 1;
	
 //Symbolic horizontal length of each stroke
 int horizontalLength = 30;

 int verticalLength = 10;




 String  NwTree = "";



 boolean print = false;



 //This is our encoding for a tree in our binary notation:
 //We use a vector of strings:
 //Even entries are the paths in the tree  of the nodes,
 //odd ones are the strings, names or contents.



 //Y-coordinates of points where strings are drawn
 int[] yCoordinate = new int[1000];
 int maxValueOfY = 0;
  int minValueOfY = 100000;


 //Set up the instructions.
 JLabel instruct1; 
 JLabel instruct2;
 JLabel instruct3;

public DrawBiTree18() 
{
 super(new BorderLayout());

     
 JPanel instructionPanel = new JPanel(new GridLayout(0,1));
       

 //Set up the instructions.
 instruct1 = new JLabel(title1);
 instruct2 = new JLabel(title2);
 instruct3 = new JLabel(title3);

 instructionPanel.setFocusable(true);
 instructionPanel.add(instruct1);
 instructionPanel.add(instruct2);
 instructionPanel.add(instruct3);

 //Set up the drawing area.
 drawingPane38 = new drawingPane38();
 drawingPane38.setBackground(Color.white);

 //Put the drawing area in a scroll pane.
 JScrollPane scroller = new JScrollPane(drawingPane38);
 scroller.setPreferredSize(new Dimension(500,900));
 scroller.revalidate();

 //Lay out this demo.
 add(instructionPanel, BorderLayout.PAGE_END);
 add(scroller, BorderLayout.CENTER);
}


//=========================================
//============= Painting class ============
//=========================================


/** The component inside the scroll pane. */
 class drawingPane38 extends JPanel 
{
 private static final long serialVersionUID = 1L;
	   
 //Initialization of coordinates
 int centerX = 1;
 int centerY = 1;
 Dimension center = new Dimension(centerX,centerY); 
 //Coordinates of the origin
 int x00;
 int y00;
 //Target coordinates
 int x11 ;
 int y11;
 //Horizontal and basic vertical strokes
 int H;
 int V;
 //Addresses are interpreted as paths	
 String path = "";
 char direction;
 int xMax = 0;
 //Repainting cause some coordinates to change:
 //work with the first ones
 boolean first = true;


//Calculate the size of a box to hold the tree
//Scrolling controls are managed
private void sizeOfTree(Graphics2D g2d)
{
 //Calculate the dimension of the strings
 //All strings have equal size.

 //Get   font in use
 Font font = g2d.getFont();
 //get metrics from the graphics
 FontMetrics metrics = g2d.getFontMetrics(font);
 //get the advance of my text in this font
 //and render context
 int lengthText = metrics.stringWidth("Aa") + 2;
 //get the height of a line of text in this
 //font and render context (some padding is included)
 int hgt = metrics.getHeight()  + 8;
 if (print)
  System.out.println("height of letters = " + hgt);

 //Horizontal and basic vertical strokes
 H = scale * horizontalLength;
 V = scale * verticalLength;

 horizontalSize = lengthText +
              (int) (H *  numberOfLeaves);
 verticalSize = 
		 (int) (hgt *   numberOfLeaves);


 //area = size of the graphic (width, height)
 Dimension area = 
		 new Dimension(horizontalSize,20*verticalSize);
 //area is set as the preferred size of the graphic
 drawingPane38.setPreferredSize(area); 
 //The graphic demands a space for itself:
 //if the alloted space is not enough, 
 //turn on scrolling bars.
 drawingPane38.revalidate();
 verticalLength = 2*verticalSize;
 //Coordinates of the origin
 x00 =    scale*centerX;
 y00 =   scale* centerY + verticalSize  - 600;
 //Target coordinates
 x11 =  x00 + (int) ( scale*horizontalLength);
 y11 =  y00;

}



//Tree is drawn
private void DrawTree3(Graphics2D g2d, Dimension center)
{	
//The tree is drawn:
//A path from the origin of the tree to
//each leaf is drawn
 if (print)
 System.out.println("LastUsedPosition at DrawTree3=  " 
                              + lastUsedPosition);
 for(int i = 0; i < lastUsedPosition; i = i+2)
  {
	//Addresses will be used as paths to  be drawn
	path = Record[i]; 
	if (print)
	System.out.println( i + " Path = " + path);
	//Reinitialization
	Dimension newCenter = 
			new Dimension(center.width,center.height);      
	V = scale * verticalLength;
	//Path is interpreted
	for(int j = 0; j < path.length(); j++)
	 {
	  direction = path.charAt(j);
	  int x0 =  newCenter .width;
	  int y0 =  newCenter .height;
	  int x1 =  x0;
	  int y1;
	  if (direction == '+') //go up
	   {      		
	    y1 =  y0 + V;       		
	    newCenter.height = newCenter.height + V;
	   }	
	   else // if direction == '-'  =  go down	 
	   {  
	     y1 =  y0 - V;        	
	     newCenter.height = newCenter.height - V;
	   } //End of if + else      
	  //Draw vertical stroke
	  g2d.drawLine(x0, verticalSize -y0, x1,
	    				verticalSize- y1);
	  //Draw horizontal stroke
	  g2d.drawLine(x1, verticalSize - y1, 
	    		   x1 + H,  verticalSize - y1);
	  if (verticalSize - y1 > maxValueOfY)
	    maxValueOfY = verticalSize - y1;
	  if (verticalSize - y1 < minValueOfY)
		    minValueOfY = verticalSize - y1;
	  //Move center       	
	  newCenter.width = newCenter.width + H ;   
	  //Vertical stroke is divided in two
	  V = V/2; 
	  //xMax = left extreme of the graphic
	  if (first)
	     if (xMax < x1 + H) xMax = x1 + H;
	  
	 }//End of for j        	            
   }  //End of for: numberOfLeaves 

 //System.out.println("xMax = " + xMax);
}



//Strings are drawn
private void drawStrings(Graphics2D g2d, Dimension center)
{
 int counter = 0;	
 //Seed is not  drawn: we lack one
 //g2d.drawString(seed,x00,verticalSize - y00);
 yCoordinate[counter] = verticalSize - y00;
 //counter++;
 //Strings are drawn:   
 for(int i = 0; i < lastUsedPosition; i = i+2)
 {
  //Addresses will be used as paths to the places
  //where strings must be drawn
  path = Record[i];
  String s = Record[i+1]; 
  //Reinitialization
  Dimension target =  
		  new Dimension(center.width,center.height);
  V = scale * verticalLength;
  //Path is interpreted
  for(int j = 0; j < path.length(); j++)
	{
	 direction = path.charAt(j);
	 if (direction == '+') //go up
	 target.height = target.height + V;    	 
	 else // if direction == '-'  =  go down	 
	 target.height = target.height - V;
	 target.width = target.width + H ;
	 V = V/2;    	
	}//End of for j
	//String is drawn
	g2d.drawString(s, target.width + 1, 
	                      verticalSize - target.height-1); 
	//y-coordinates are recorded
	yCoordinate[counter] = verticalSize - target.height-4;
	 counter++;
  }    
} 	


//Basic routine of the drawing procedure
private void drawBasic(Graphics2D g2d)
{
 g2d.setColor(Color.red);
 /*
 //Stem is drawn: our trees is artificially rooted.  
 g2d.drawLine(x00, verticalSize - y00, 
	  		   x11,  verticalSize - y11);
*/
 sizeOfTree(g2d);
 
 //targetTree is drawn when it exists
 if (inputMatrix) xMax = 0;
  else
  {
   decoderLengths(targetTree);

   //Absolute distance from the associated distance matrix
   //of the tree
   //to target matrix is evaluated
   //d = absDistance(numberOfLeaves,targetMatrix, M);
   associatedMatrix(targetTree); 
   center = new Dimension(x11,y11);
   DrawTree3(g2d, center);
   center = new Dimension(x00,y00);
   drawStrings(g2d,center);
   first = false;
  }
 
 
 //bestTree is drawn:
 //Absolute distance from the associated distance matrix
 //of the tree
 //to target matrix is evaluated
 //d = absDistance(numberOfLeaves,targetMatrix, M);
 
 
 if (inputMatrix) 
 {
  decoderLengths(bestTree);
  associatedMatrix(bestTree); 
  center = new Dimension(x11,y11);
  DrawTree3(g2d, center);
  center = new Dimension(x00,y00);
  drawStrings(g2d,center);	 
 }
 else
 {
  decoderLengths(bestTree);
  associatedMatrix(bestTree); 

  center = new Dimension(xMax + 30,y11);
  DrawTree3(g2d, center);
  center = new Dimension(
		 xMax - (int) ( scale*horizontalLength) + 30,y00);
 drawStrings(g2d,center);
 }
 
}

//Graphical  master	: it is called automatically 
//at starting and by repaint()
protected void paintComponent(Graphics g) 
{
 super.paintComponent(g);
 Graphics2D g2d = (Graphics2D)g;
 //Font to draw strings
 Font font = new Font("Arial", Font.PLAIN, 12);  
 g2d.setFont(font);
 g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
		  RenderingHints.VALUE_ANTIALIAS_ON);  
  drawBasic(g2d);   		
}


}//End of double inner class drawingPane38



/**
* Create the GUI and show it.  For thread safety,
* this method should be invoked from the
* event-dispatching thread.
*/
 void createAndShowGUI() 
{
 //Create and set up the window.
 JFrame frame = 
   new JFrame(titleFrame);
 frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);

 //Create and set up the content pane.
 //Constructor of the main application is used.
 JComponent newContentPane = new DrawBiTree18();
 //content panes must be opaque
 newContentPane.setOpaque(true); 
 frame.setContentPane(newContentPane);

 //Display the window.
 frame.pack();
 frame.setVisible(true);
}

}//End of  inner class DrawBiTree18

}//End of main class K65 EvolFull10

